# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import datetime
import unittest2

from sqlalchemy import Column, Integer, ForeignKey
from stalker.models.mixins import ScheduleMixin
from stalker.db.session import DBSession
from stalker.models.entity import SimpleEntity

from stalker import config
defaults = config.Config()


class SchedMixFooMixedInClass(SimpleEntity, ScheduleMixin):
    """a class which derives from another which has and __init__ already
    """
    __tablename__ = "SchedMixFooMixedInClasses"
    __mapper_args__ = {"polymorphic_identity": "SchedMixFooMixedInClass"}
    schedMixFooMixedInClass_id = Column("id", Integer,
                                        ForeignKey("SimpleEntities.id"),
                                        primary_key=True)

    def __init__(self, **kwargs):
        super(SchedMixFooMixedInClass, self).__init__(**kwargs)
        ScheduleMixin.__init__(self, **kwargs)


class ScheduleMixinTester(unittest2.TestCase):
    """Tests the ScheduleMixin
    """

    def setUp(self):
        """setup the test
        """
        # create mock objects
        self.start = datetime.datetime(2013, 3, 22, 15, 15)
        self.end = self.start + datetime.timedelta(days=20)
        self.duration = datetime.timedelta(days=10)

        self.kwargs = {
            'name': 'Test Schedule Mixin',
            'description': 'This is a simple entity object for testing '
                           'ScheduleMixin',
            'start': self.start,
            'end': self.end,
            'duration': self.duration,
            'timing_resolution': datetime.timedelta(minutes=30)
        }

        self.mock_foo_obj = SchedMixFooMixedInClass(**self.kwargs)

    def tearDown(self):
        """clean up the test
        """
        DBSession.remove()

    def test_start_argument_is_not_a_date_object(self):
        """testing if defaults will be used for the start attribute when
        the start is given as something other than a datetime.datetime object
        """
        test_values = [1, 1.2, "str", ["a", "date"]]

        for test_value in test_values:
            self.kwargs["start"] = test_value

            new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)

            self.assertEqual(new_foo_obj.start,
                             new_foo_obj.end - new_foo_obj.duration)
            
            # the values are rounded can not check anymore
            #self.assertEqual(new_foo_obj.end, self.kwargs["end"])
            #self.assertEqual(new_foo_obj.duration, self.kwargs["duration"])
    
    def test_start_attribute_is_not_a_date_object(self):
        """testing if the defaults will be used when trying to set the
        start attribute to something other than a datetime.datetime object
        """
        test_values = [1, 1.2, "str", ["a", "date"]]
        
        for test_value in test_values:
            end = self.mock_foo_obj.end
            duration = self.mock_foo_obj.duration

            self.mock_foo_obj.start = test_value

            self.assertEqual(
                self.mock_foo_obj.start,
                self.mock_foo_obj.end - self.mock_foo_obj.duration
            )

            # check if we still have the same end
            self.assertEqual(self.mock_foo_obj.end, end)

            # check if we still have the same duration
            self.assertEqual(self.mock_foo_obj.duration, duration)
    
    def test_start_attribute_is_set_to_None_use_the_default_value(self):
        """testing if setting the start attribute to None will update the
        start to today
        """
        self.mock_foo_obj.start = None
        self.assertEqual(self.mock_foo_obj.start,
                         datetime.datetime(2013, 3, 22, 15, 30))
        self.assertIsInstance(self.mock_foo_obj.start, datetime.datetime)
    
    def test_start_attribute_works_properly(self):
        """testing if the start properly is working properly
        """
        test_value = datetime.datetime(year=2011, month=1, day=1)
        self.mock_foo_obj.start = test_value
        self.assertEqual(self.mock_foo_obj.start, test_value)
    
    def test_end_argument_is_not_a_date_object(self):
        """testing if default values will be for the end attribute used when
        trying to set the due date something other than a datetime.datetime
        object
        """
        test_values = [1, 1.2, "str", ["a", "date"],
                       datetime.timedelta(days=100)]
        
        for test_value in test_values:
            self.kwargs["end"] = test_value
            new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)

            self.assertEqual(new_foo_obj.end,
                             new_foo_obj.start + new_foo_obj.duration)
    
    def test_end_attribute_is_not_a_date_object(self):
        """testing if default values will be used for the end attribute
        when trying to set the end attribute to something other than a
        datetime.datetime object
        """
        test_values = [1, 1.2, "str", ["a", "date"],
                       datetime.timedelta(days=100)]
        
        for test_value in test_values:
            self.mock_foo_obj.end = test_value
            
            self.assertEqual(self.mock_foo_obj.end,
                             self.mock_foo_obj.start +\
                             self.mock_foo_obj.duration)
    
    def test_end_argument_is_tried_to_set_to_a_time_before_start(
    self):
        """testing if end attribute will be updated to
        start + duration when the end argument is given as a value
        which is a date before start
        """

        self.kwargs["end"] = self.kwargs["start"] -\
                                  datetime.timedelta(days=10)

        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)

        self.assertEqual(new_foo_obj.end,
                         new_foo_obj.start + new_foo_obj.duration)
    
    def test_end_attribute_is_tried_to_set_to_a_time_before_start(
    self):
        """testing if a the due attribute is re calculated from the start
        and duration attributes when the end is tried to be set to a date
        before start
        """
        new_end = self.mock_foo_obj.start - datetime.timedelta(days=10)
        self.mock_foo_obj.end = new_end
        self.assertEqual(self.mock_foo_obj.end,
                         self.mock_foo_obj.start +\
                         self.mock_foo_obj.duration)
    
    def test_end_attribute_is_shifted_when_start_passes_it(self):
        """testing if end attribute will be shifted when the start
        attribute passes it
        """
        time_delta = self.mock_foo_obj.end - self.mock_foo_obj.start
        self.mock_foo_obj.start += 2 * time_delta

        self.assertEqual(
            self.mock_foo_obj.end - self.mock_foo_obj.start,
            time_delta
        )

    def test_duration_argument_is_not_an_instance_of_timedelta_no_problem_if_start_and_end_is_present(self):
        """testing if no error will be raised when the duration argument is not
        an instance of datetime.datetime class when both of the start and end
        arguments are present.
        """
        test_values = [1, 1.2, "10", "10 days"]
        
        # no problem if there are start and end arguments
        for test_value in test_values:
            self.kwargs["duration"] = test_value
            new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)

            # check the value
            self.assertEqual(new_foo_obj.duration,
                             new_foo_obj.end - new_foo_obj.start)

    def test_duration_argument_is_not_an_instance_of_date_when_start_argument_is_missing(self):
        """testing if defaults for the duration attribute will be used  when
        the duration argument is not an instance of datetime.datetime class
        when start argument is also missing
        """
        test_values = [1, 1.2, "10", "10 days"]
        self.kwargs.pop("start")
        for test_value in test_values:
            self.kwargs["duration"] = test_value
            new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)

            self.assertEqual(new_foo_obj.duration,
                             defaults.task_duration)

    def test_duration_argument_is_not_an_instance_of_date_when_end_argument_is_missing(self):
        """testing if the defaults for the duration attribute will be used when
        the duration argument is not an instance of datetime.datetime class and
        when end argument is also missing
        """
        test_values = [1, 1.2, "10", "10 days"]
        self.kwargs.pop("end")
        for test_value in test_values:
            self.kwargs["duration"] = test_value
            new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
            self.assertEqual(new_foo_obj.duration,
                             defaults.task_duration)

    def test_duration_attribute_is_calculated_correctly(self):
        """testing if the duration attribute is calculated correctly
        """
        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)
        new_foo_entity.start = datetime.datetime(2013, 3, 22, 15, 0)
        new_foo_entity.end = new_foo_entity.start +\
                                  datetime.timedelta(201)

        self.assertEqual(new_foo_entity.duration, datetime.timedelta(201))

    def test_duration_attribute_is_set_to_not_an_instance_of_timedelta(self):
        """testing if duration attribute reset to a calculated value when it is
        set to something other than a datetime.timedelta instance
        """
        test_values = [1, 1.2, "10", "10 days"]

        # no problem if there are start and end arguments
        for test_value in test_values:
            self.mock_foo_obj.duration = test_value

            # check the value
            self.assertEqual(self.mock_foo_obj.duration,
                             self.mock_foo_obj.end -\
                             self.mock_foo_obj.start)

    def test_duration_attribute_expands_then_end_shifts(self):
        """testing if duration attribute is expanded then the end
        attribute is shifted
        """
        end = self.mock_foo_obj.end
        start = self.mock_foo_obj.start
        duration = self.mock_foo_obj.duration

        # change the duration
        new_duration = duration * 10
        self.mock_foo_obj.duration = new_duration

        # duration expanded
        self.assertEqual(self.mock_foo_obj.duration, new_duration)

        # start is not changed
        self.assertEqual(self.mock_foo_obj.start, start)

        # end is postponed
        self.assertEqual(self.mock_foo_obj.end, start + new_duration)

    def test_duration_attribute_contracts_then_end_shifts_back(self):
        """testing if duration attribute is contracted then the end
        attribute is shifted back
        """
        end = self.mock_foo_obj.end
        start = self.mock_foo_obj.start
        duration = self.mock_foo_obj.duration

        # change the duration
        new_duration = duration / 2
        self.mock_foo_obj.duration = new_duration

        # duration expanded
        self.assertEqual(self.mock_foo_obj.duration, new_duration)

        # start is not changed
        self.assertEqual(self.mock_foo_obj.start, start)

        # end is postponed
        self.assertEqual(self.mock_foo_obj.end, start + new_duration)

    def test_duration_is_a_negative_timedelta(self):
        """testing if the duration is a negative timedelta will set the
        duration to 1 days
        """
        start = self.mock_foo_obj.start
        self.mock_foo_obj.duration = datetime.timedelta(-10)
        self.assertEqual(self.mock_foo_obj.duration, datetime.timedelta(1))
        self.assertEqual(self.mock_foo_obj.start, start)

    def test_init_all_parameters_skipped(self):
        """testing if the attributes are initialized to:
        
        start = datetime.datetime.now()
        duration = stalker.config.Config.task_duration
        end = start + duration
        """
        #self.fail("test is not implemented yet")
        self.kwargs.pop("start")
        self.kwargs.pop("end")
        self.kwargs.pop("duration")
        
        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)
        
        self.assertIsInstance(new_foo_entity.start, datetime.datetime)
        # can not check for start, just don't want to strugle with the round
        # thing
        #self.assertEqual(new_foo_entity.start,
        #                 datetime.datetime(2013, 3, 22, 15, 30))
        self.assertEqual(new_foo_entity.duration,
                         defaults.task_duration)
        self.assertEqual(new_foo_entity.end,
                         new_foo_entity.start + new_foo_entity.duration)
    
    def test_init_only_start_argument_is_given(self):
        """testing if the attributes are initialized to:
        
        duration = stalker.config.Config.default_task_duration
        end = start + duration
        """
        self.kwargs.pop("end")
        self.kwargs.pop("duration")

        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)

        self.assertEqual(new_foo_entity.duration, defaults.task_duration)
        self.assertEqual(new_foo_entity.end,
                         new_foo_entity.start + new_foo_entity.duration)
    
    def test_init_start_and_end_argument_is_given(self):
        """testing if the attributes are initialized to:
        
        duration = end - start
        """
        self.kwargs.pop("duration")
        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)
        self.assertEqual(new_foo_entity.duration,
                         new_foo_entity.end - new_foo_entity.start)
    
    def test_init_start_and_duration_argument_is_given(self):
        """testing if the attributes are initialized to:
        
        end = start + duration
        """

        self.kwargs.pop("end")

        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)

        self.assertEqual(new_foo_entity.end,
                         new_foo_entity.start + new_foo_entity.duration)
    
    def test_init_all_arguments_are_given(self):
        """testing if the attributes are initialized to:
        
        duration = end - start
        """
        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)
        self.assertEqual(new_foo_entity.duration,
                         new_foo_entity.end - new_foo_entity.start)
    
    def test_init_end_and_duration_argument_is_given(self):
        """testing if the attributes are initialized to:
        
        start = end - duration
        """
        self.kwargs.pop("start")
        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)
        self.assertEqual(new_foo_entity.start,
                         new_foo_entity.end - new_foo_entity.duration)
    
    def test_init_only_end_argument_is_given(self):
        """testing if the attributes are initialized to:
        
        duration = stalker.config.Config.default_task_duration
        start = end - duration
        """
        self.kwargs.pop("duration")
        self.kwargs.pop("start")
        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)
        self.assertEqual(new_foo_entity.duration,
                         defaults.task_duration)
        self.assertEqual(new_foo_entity.start,
                         new_foo_entity.end - new_foo_entity.duration)
    
    def test_init_only_duration_argument_is_given(self):
        """testing if the attributes are initialized to:
        
        start = datetime.date.today()
        end = start + duration
        """
        self.kwargs.pop("end")
        self.kwargs.pop("start")
        
        new_foo_entity = SchedMixFooMixedInClass(**self.kwargs)

        # just check if it is an instance of datetime.datetime
        self.assertIsInstance(new_foo_entity.start, datetime.datetime)
        # can not check for start
        #self.assertEqual(new_foo_entity.start,
        #                 datetime.datetime(2013, 3, 22, 15, 30))
        self.assertEqual(new_foo_entity.end,
                         new_foo_entity.start + new_foo_entity.duration)
    
    def test_resolution_argument_skipped(self):
        """testing if the timing_resolution attribute will be set to the default value
        from the defaults.timing_resolution if timing_resolution argument is skipped
        """
        try:
            self.kwargs.pop('timing_resolution')
        except KeyError:
            pass
        
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertEqual(new_foo_obj.timing_resolution, defaults.timing_resolution)
    
    def test_resolution_argument_is_None(self):
        """testing if the timing_resolution attribute will be set to the default value
        from the defaults.timing_resolution if timing_resolution argument is None
        """
        self.kwargs['timing_resolution'] = None
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertEqual(new_foo_obj.timing_resolution, defaults.timing_resolution)
    
    def test_resolution_attribute_is_set_to_None(self):
        """testing if the timing_resolution attribute will be set to the default value
        from the defaults.timing_resolution if it is set to None
        """
        self.kwargs['timing_resolution'] = datetime.timedelta(minutes=5)
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        # check start conditions
        self.assertEqual(new_foo_obj.timing_resolution, self.kwargs['timing_resolution'])
        new_foo_obj.timing_resolution = None
        self.assertEqual(new_foo_obj.timing_resolution, defaults.timing_resolution)
    
    def test_resolution_argument_is_not_a_timedelta_instance(self):
        """testing if a TypeError will be raised when the timing_resolution argument
        is not a datetime.timedelta instance
        """
        self.kwargs['timing_resolution'] = 'not a timedelta instance'
        self.assertRaises(TypeError, SchedMixFooMixedInClass, **self.kwargs)
    
    def test_resolution_attribute_is_not_a_timedelta_instance(self):
        """testing if a TypeError will be raised when the timing_resolution attribute
        is not a datetime.timedelta instance
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertRaises(TypeError, setattr, new_foo_obj, 'timing_resolution',
                          'not a timedelta instance')
    
    def test_resolution_argument_is_working_properly(self):
        """testing if the timing_resolution argument value is passed to timing_resolution
        attribute correctly
        """
        self.kwargs['timing_resolution'] = datetime.timedelta(minutes=5)
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertEqual(new_foo_obj.timing_resolution, self.kwargs['timing_resolution'])
    
    def test_resolution_attribute_is_working_properly(self):
        """testing if the timing_resolution attribute is working properly
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        res = new_foo_obj
        new_res = datetime.timedelta(hours=1, minutes=30)
        self.assertNotEqual(res, new_res)
        new_foo_obj.timing_resolution = new_res
        self.assertEqual(new_foo_obj.timing_resolution, new_res)
    
    def test_start_end_and_duration_values_are_rounded_to_the_defined_resolution_argument(self):
        """testing if the start and end dates are rounded to the given
        timing_resolution
        """
        self.kwargs['start'] = datetime.datetime(2013,3,22,02,38,55,531)
        self.kwargs['end'] = datetime.datetime(2013,3,24,16,46,32,102)
        self.kwargs['timing_resolution'] = datetime.timedelta(minutes=5)
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        # check the start
        expected_start = datetime.datetime(2013,3,22,02,40)
        self.assertEqual(new_foo_obj.start, expected_start)
        # check the end
        expected_end = datetime.datetime(2013,3,24,16,45)
        self.assertEqual(new_foo_obj.end, expected_end)
        # check the duration
        self.assertEqual(new_foo_obj.duration, expected_end - expected_start)
   
    def test_computed_start_is_None_for_a_non_scheduled_class(self):
        """testing if the computed_start attribute is None for a non scheduled
        class
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertIsNone(new_foo_obj.computed_start, None)
    
    def test_computed_end_is_None_for_a_non_scheduled_class(self):
        """testing if the computed_end attribute is None for a non scheduled
        class
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertIsNone(new_foo_obj.computed_end, None)
    
    def test_computed_duration_attribute_is_None_if_there_is_no_computed_start_and_no_computed_end(self):
        """testing if the computed_start attribute is None if there is no
        computed_start and no computed_end
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        new_foo_obj.computed_start = None
        new_foo_obj.computed_end = None
        self.assertIsNone(new_foo_obj.computed_duration)
    
    def test_computed_duration_attribute_is_None_if_there_is_computed_start_but_no_computed_end(self):
        """testing if the computed_start attribute is None if there is
        computed_start but no computed_end
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        new_foo_obj.computed_start = datetime.datetime.now()
        new_foo_obj.computed_end = None
        self.assertIsNone(new_foo_obj.computed_duration)
    
    def test_computed_duration_attribute_is_None_if_there_is_no_computed_start_but_computed_end(self):
        """testing if the computed_start attribute is None if there is no
        computed_start but computed_end
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        new_foo_obj.computed_start = None
        new_foo_obj.computed_end = datetime.datetime.now()
        self.assertIsNone(new_foo_obj.computed_duration)
    
    def test_computed_duration_attribute_is_calculated_correctly_when_there_are_both_computed_start_and_computed_end(self):
        """testing if the computed_duration is calculated correctly when there
        are both computed_start and computed_end values
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        new_foo_obj.computed_start = datetime.datetime.now()
        new_foo_obj.computed_end =  new_foo_obj.computed_start + \
                                    datetime.timedelta(12)
        self.assertEqual(new_foo_obj.computed_duration, datetime.timedelta(12))
    
    def test_computed_duration_is_read_only(self):
        """testing if the computed_duration attribute is read-only
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertRaises(AttributeError, setattr, new_foo_obj,
                          'computed_duration', datetime.timedelta(10))

    def test_total_seconds_attribute_is_read_only(self):
        """testing if the total_seconds is read only
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertRaises(
            AttributeError,
            setattr,
            new_foo_obj,
            'total_seconds',
            234234
        )

    def test_total_seconds_attribute_is_working_properly(self):
        """testing is the total_seconds is read only
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        new_foo_obj.start = datetime.datetime(2013, 5, 31, 10, 0)
        new_foo_obj.end = datetime.datetime(2013, 5, 31, 18, 0)
        self.assertEqual(
            new_foo_obj.total_seconds,
            8 * 60 * 60
        )

    def test_computed_total_seconds_attribute_is_read_only(self):
        """testing if the computed_total_seconds is read only
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        self.assertRaises(
            AttributeError,
            setattr,
            new_foo_obj,
            'computed_total_seconds',
            234234
        )

    def test_computed_total_seconds_attribute_is_working_properly(self):
        """testing is the computed_total_seconds is read only
        """
        new_foo_obj = SchedMixFooMixedInClass(**self.kwargs)
        new_foo_obj.computed_start = datetime.datetime(2013, 5, 31, 10, 0)
        new_foo_obj.computed_end = datetime.datetime(2013, 5, 31, 18, 0)
        self.assertEqual(
            new_foo_obj.computed_total_seconds,
            8 * 60 * 60
        )



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
import os

import unittest2
from stalker import Type, Link


class LinkTester(unittest2.TestCase):
    """tests the :class:`stalker.models.link.Link` class
    """

    def setUp(self):
        """setup the test
        """
        # create a mock LinkType object
        self.test_link_type1 = Type(
            name='Test Type 1',
            code='test type1',
            target_entity_type=Link,
        )
        self.test_link_type2 = Type(
            name='Test Type 2',
            code='test type2',
            target_entity_type=Link,
        )

        self.kwargs = {
            'name': 'An Image Link',
            'full_path': 'C:/A_NEW_PROJECT/td/dsdf/22-fdfffsd-32342-dsf2332-dsfd-3.exr',
            'original_filename': 'this_is_an_image.jpg',
            'type': self.test_link_type1
        }

        self.test_link = Link(**self.kwargs)

    def test___auto_name__class_attribute_is_set_to_True(self):
        """testing if the __auto_name__ class attribute is set to False for
        Link class
        """
        self.assertTrue(Link.__auto_name__)

    def test_full_path_argument_is_None(self):
        """testing if setting the full_path argument to None will set the
        full_path attribute to an empty string
        """
        self.kwargs['full_path'] = None
        new_link = Link(**self.kwargs)
        self.assertEqual(new_link.full_path, '')

    def test_full_path_attribute_is_set_to_None(self):
        """testing if setting the full_path attribute to None will set its
        value to an empty string
        """
        self.test_link.full_path = ''

    def test_full_path_argument_is_empty_an_empty_string(self):
        """testing if setting the full_path attribute will be an empty string
        if full_path argument is an empty string
        """
        self.kwargs['full_path'] = ''
        new_link = Link(**self.kwargs)
        self.assertEqual(new_link.full_path, '')

    def test_full_path_attribute_is_set_to_empty_string(self):
        """testing if setting the full_path attribute value will be an empty
        string if it is set to an empty string
        """
        self.test_link.full_path = ''
        self.assertEqual(self.test_link.full_path, '')

    def test_full_path_argument_is_not_a_string(self):
        """testing if a TypeError will be raised when the full_path argument is
        not a string or unicode instance
        """
        test_values = [1, 1.1, ['a path'], {'a': 'path'}]

        for test_value in test_values:
            self.kwargs["full_path"] = test_value
            self.assertRaises(TypeError, Link, **self.kwargs)

    def test_full_path_attribute_is_not_a_string(self):
        """testing if a TypeError will be raised when the full_path attribute
        is not a string or unicode instance
        """
        test_values = [1, 1.1, ['a path'], {'a': 'path'}]

        for test_value in test_values:
            self.assertRaises(
                TypeError,
                setattr,
                self.test_link,
                'full_path',
                test_value
            )

    def test_full_path_windows_to_other_conversion(self):
        """testing if the full_path is stored in internal format
        """
        windows_path = 'M:\\path\\to\\object'
        expected_result = 'M:/path/to/object'

        self.test_link.full_path = windows_path
        self.assertEqual(self.test_link.full_path, expected_result)


    def test_original_filename_argument_is_None(self):
        """testing if the original_filename attribute value will be set to the
        filename part of the full_path attribute if the original_filename
        argument is None
        """
        self.kwargs['original_filename'] = None
        new_link = Link(**self.kwargs)
        filename = os.path.basename(new_link.full_path)
        self.assertEqual(new_link.original_filename, filename)

    def test_original_filename_attribute_is_set_to_None(self):
        """testing if the original_filename attribute value will be equal to
        the filename part of the full_path if it is set to None
        """
        self.test_link.original_filename = None
        filename = os.path.basename(self.test_link.full_path)
        self.assertEqual(self.test_link.original_filename, filename)

    def test_original_filename_argument_is_empty_string(self):
        """testing if setting the original_filename argument to an empty string
        will set the original_filename attribute value to the filename part of
        the full_path
        """
        self.kwargs['original_filename'] = ''
        new_link = Link(**self.kwargs)
        filename = os.path.basename(new_link.full_path)
        self.assertEqual(new_link.original_filename, filename)

    def test_original_filename_attribute_set_to_empty_string(self):
        """testing if setting the original_filename attribute to an empty
        string will set its value to filename part of the full_path attribute
        """
        self.test_link.original_filename = ''
        filename = os.path.basename(self.test_link.full_path)
        self.assertEqual(self.test_link.original_filename, filename)

    def test_original_filename_argument_accepts_string_or_unicode_only(self):
        """testing if original_filename argument accepts string or unicode only
        and raises TypeError for other types
        """
        test_values = [1, 1.1, ['a original_filename'],
                       {'a': 'original_filename'}]
        for test_value in test_values:
            self.kwargs['original_filename'] = test_value
            self.assertRaises(TypeError, Link, **self.kwargs)

    def test_original_filename_attribute_accepts_string_or_unicode_only(self):
        """testing if original_filename attribute accepts string or unicode
        only and raises TypeError for other types
        """
        test_values = [1, 1.1, ['a original_filename'],
                       {'a': 'original_filename'}]
        for test_value in test_values:
            self.assertRaises(
                TypeError,
                setattr,
                self.test_link,
                'original_filename',
                test_value
            )

    def test_original_filename_argument_is_working_properly(self):
        """testing if the original_filename argument is working properly
        """
        self.assertEqual(
            self.kwargs['original_filename'],
            self.test_link.original_filename
        )

    def test_original_filename_attribute_is_working_properly(self):
        """testing if the original_filename attribute is working properly
        """
        new_value = 'this_is_the_original_filename.jpg'
        self.assertNotEqual(
            self.test_link.original_filename,
            new_value
        )
        self.test_link.original_filename = new_value
        self.assertEqual(
            self.test_link.original_filename,
            new_value
        )

    def test_equality_of_two_links(self):
        """testing the equality operator
        """
        # with same parameters
        mock_link1 = Link(**self.kwargs)
        self.assertTrue(self.test_link == mock_link1)

        # with different parameters
        self.kwargs['type'] = self.test_link_type2
        mock_link2 = Link(**self.kwargs)

        self.assertFalse(self.test_link == mock_link2)

    def test_inequality_of_two_links(self):
        """testing the inequality operator
        """
        # with same parameters
        mock_link1 = Link(**self.kwargs)
        self.assertTrue(self.test_link == mock_link1)

        # with different parameters
        self.kwargs['type'] = self.test_link_type2
        mock_link2 = Link(**self.kwargs)

        self.assertFalse(self.test_link != mock_link1)
        self.assertTrue(self.test_link != mock_link2)

    def test_plural_class_name(self):
        """testing the plural name of Link class
        """
        self.assertEqual(self.test_link.plural_class_name, "Links")

        # def test___strictly_typed___is_True(self):
        # """testing if the __strictly_typed__ class attribute is True
        # """
        # self.assertEqual(Link.__strictly_typed__, True)

    def test_path_attribute_is_set_to_None(self):
        """testing if a TypeError will be raised when the path attribute is set
        to None
        """
        self.assertRaises(TypeError, setattr, self.test_link, 'path', None)

    def test_path_attribute_is_set_to_empty_string(self):
        """testing if a ValueError will be raised when the path attribute is
        set to an empty string
        """
        self.assertRaises(ValueError, setattr, self.test_link, 'path', '')

    def test_path_attribute_is_set_to_a_value_other_then_string(self):
        """testing if a TypeError will be raised when the path attribute is set
        to a value other than string
        """
        self.assertRaises(TypeError, setattr, self.test_link, 'path', 1)

    def test_path_attribute_value_comes_from_full_path(self):
        """testing if the path attribute value is calculated from the full_path
        attribute
        """
        self.assertEqual(
            self.test_link.path,
            'C:/A_NEW_PROJECT/td/dsdf',
        )

    def test_path_attribute_updates_the_full_path_attribute(self):
        """testing if the path attribute is updating the full_path attribute
        properly
        """
        test_value = '/mnt/some/new/path'
        expected_full_path = '/mnt/some/new/path/22-fdfffsd-32342-dsf2332-dsfd-3.exr'

        self.assertNotEqual(self.test_link.path, test_value)
        self.test_link.path = test_value
        self.assertEqual(self.test_link.path, test_value)
        self.assertEqual(self.test_link.full_path, expected_full_path)

    def test_filename_attribute_is_set_to_None(self):
        """testing if the filename attribute will be an empty string if it is
        set a None
        """
        self.test_link.filename = None
        self.assertEqual(self.test_link.filename, '')

    def test_filename_attribute_is_set_to_a_value_other_then_string(self):
        """testing if a TypeError will be raised when the filename attribute is
        set to a value other than string
        """
        self.assertRaises(TypeError, setattr, self.test_link, 'filename', 3)

    def test_filename_attribute_is_set_to_empty_string(self):
        """testing if the filename value can be set to an empty string
        """
        self.test_link.filename = ''
        self.assertEqual(self.test_link.filename, '')

    def test_filename_attribute_value_comes_from_full_path(self):
        """testing if the filename attribute value is calculated from the
        full_path attribute
        """
        self.assertEqual(
            self.test_link.filename,
            '22-fdfffsd-32342-dsf2332-dsfd-3.exr'
        )

    def test_filename_attribute_updates_the_full_path_attribute(self):
        """testing if the filename attribute is updating the full_path
        attribute properly
        """
        test_value = 'new_filename.tif'
        self.assertNotEqual(self.test_link.filename, test_value)
        self.test_link.filename = test_value
        self.assertEqual(self.test_link.filename, test_value)
        self.assertEqual(self.test_link.full_path,
                         'C:/A_NEW_PROJECT/td/dsdf/new_filename.tif')
    
    def test_filename_attribute_changes_also_the_extension(self):
        """testing if the filename attribute also changes the extension
        attribute
        """
        self.assertNotEqual(self.test_link.extension, '.an_extension')
        self.test_link.filename = 'some_filename_and.an_extension'
        self.assertEqual(self.test_link.extension, '.an_extension')

    def test_extension_attribute_is_set_to_None(self):
        """testing if the extension will be an empty string if it is set to
        None
        """
        self.test_link.extension = None
        self.assertEqual(self.test_link.extension, '')

    def test_extension_attribute_is_set_to_empty_string(self):
        """testing if the value of the extension can be set to an empty string
        """
        self.test_link.extension = ''
        self.assertEqual(self.test_link.extension, '')

    def test_extension_attribute_is_set_to_a_value_other_then_string(self):
        """testing if a TypeError will be raised when the extension attribute
        is set to a value other than string
        """
        self.assertRaises(TypeError, setattr, self.test_link, 'extension', 123)

    def test_extension_attribute_value_comes_from_full_path(self):
        """testing if the extension attribute value is calculated from the
        full_path attribute
        """
        self.assertEqual(self.test_link.extension, '.exr')

    def test_extension_attribute_updates_the_full_path_attribute(self):
        """testing if the extension attribute is updating the full_path
        attribute properly
        """
        test_value = '.iff'
        self.assertNotEqual(self.test_link.extension, test_value)
        self.test_link.extension = test_value
        self.assertEqual(self.test_link.extension, test_value)
        self.assertEqual(
            self.test_link.full_path,
            'C:/A_NEW_PROJECT/td/dsdf/22-fdfffsd-32342-dsf2332-dsfd-3.iff'
        )

    def test_extension_attribute_updates_the_full_path_attribute_with_the_dot(self):
        """testing if the extension attribute is updating the full_path
        attribute properly even if a dot is not given as the first character
        """
        test_value = 'iff'
        expected_value = '.iff'
        self.assertNotEqual(self.test_link.extension, test_value)
        self.test_link.extension = test_value
        self.assertEqual(self.test_link.extension, expected_value)
        self.assertEqual(
            self.test_link.full_path,
            'C:/A_NEW_PROJECT/td/dsdf/22-fdfffsd-32342-dsf2332-dsfd-3.iff'
        )

    def test_extension_attribute_will_also_change_the_filename_attribute(self):
        """testing if the extension attribute is updating the filename
        attribute
        """
        test_value = '.targa'
        expected_value = '22-fdfffsd-32342-dsf2332-dsfd-3.targa'
        self.assertNotEqual(self.test_link.filename, expected_value)
        self.test_link.extension = test_value
        self.assertEqual(self.test_link.filename, expected_value)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import datetime
import re
import uuid
from sqlalchemy import Table, Column, Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import relationship, validates, reconstructor

import stalker
from stalker.db import Base

from stalker.log import logging_level
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


class SimpleEntity(Base):
    """The base class of all the others

    The ``SimpleEntity`` is the starting point of the Stalker Object Model, it
    starts by adding the basic information about an entity which are
    :attr:`.name`, :attr:`.description`, the audit information like
    :attr:`.created_by`, :attr:`.updated_by`, :attr:`.date_created`,
    :attr:`.date_updated` and a couple of naming attributes like
    :attr:`.nice_name` and last but not least the :attr:`.type` attribute which
    is very important for entities that needs a type.

    .. note::

       For derived classes if the
       :attr:`~stalker.models.entity.SimpleEntity.type` needed to be
       specifically specified, that is it can not be None or nothing else then
       a :class:`~stalker.models.type.Type` instance, set the
       ``strictly_typed`` class attribute to True::

           class NewClass(SimpleEntity):
               __strictly_typed__ = True

       This will ensure that the derived class always have a proper
       :attr:`~stalker.models.entity.SimpleEntity.type` attribute and can not
       be initialized without one.

    Two SimpleEntities considered to be equal if they have the same
    :attr:`.name`, the
    other attributes doesn't matter.

    .. versionadded:: 0.2.0
       Name attribute can be skipped. Starting from version 0.2.0 the ``name``
       attribute can be skipped. For derived classes use the ``__auto_name__``
       class attribute to control auto naming behaviour.

    :param string name: A string or unicode value that holds the name of this
      entity.  It should not contain any white space at the beginning and at
      the end of the string. Valid characters are [a-zA-Z0-9\_/S].

      Advanced::

        For classes derived from the SimpleEntity, if an automatic name is
        desired, the ``__auto_name__`` class attribute can be set to True. Then
        Stalker will automatically generate a uuid4 sequence for the name
        attribute.

    :param str description: A string or unicode attribute that holds the
      description of this entity object, it could be an empty string, and it
      could not again have white spaces at the beginning and at the end of the
      string, again any given objects will be converted to strings

    :param created_by: The :class:`~stalker.models.auth.User` who has created
      this object

    :type created_by: :class:`~stalker.models.auth.User`

    :param updated_by: The :class:`~stalker.models.auth.User` who has updated
      this object lastly. The created_by and updated_by attributes point the
      same object if this object is just created.

    :param date_created: The date that this object is created.

    :type date_created: :class:`datetime.datetime`

    :param date_updated: The date that this object is updated lastly. For newly
      created entities this is equal to date_created and the date_updated
      cannot point a date which is before date_created.

    :type date_updated: :class:`datetime.datetime`

    :param type: The type of the current SimpleEntity. Used across several
      places in Stalker. Can be None. The default value is None.

    :type type: :class:`~stalker.models.type.Type`
    """
    # auto generate name values
    __auto_name__ = True
    __strictly_typed__ = False

    # TODO: Allow the user to specify the formatting of the name attribute as a Regular Expression (name_formatter)
    __name_formatter__ = None

    __tablename__ = "SimpleEntities"
    id = Column("id", Integer, primary_key=True)

    entity_type = Column(String(128), nullable=False)
    __mapper_args__ = {
        "polymorphic_on": entity_type,
        "polymorphic_identity": "SimpleEntity"
    }

    name = Column(
        String(256),
        nullable=False,
        doc="""Name of this object"""
    )

    description = Column(
        "description",
        String,
        doc="""Description of this object."""
    )

    created_by_id = Column(
        "created_by_id",
        Integer,
        ForeignKey("Users.id", use_alter=True, name="xc"),
        doc="""The id of the :class:`~stalker.models.auth.User` who has created
        this entity."""
    )

    created_by = relationship(
        "User",
        backref="entities_created",
        primaryjoin="SimpleEntity.created_by_id==User.user_id",
        post_update=True,
        doc="""The :class:`~stalker.models.auth.User` who has created this object."""
    )

    updated_by_id = Column(
        "updated_by_id",
        Integer,
        ForeignKey("Users.id", use_alter=True, name="xu"),
        doc="""The id of the :class:`~stalker.models.auth.User` who has updated
        this entity."""
    )

    updated_by = relationship(
        "User",
        backref="entities_updated",
        primaryjoin="SimpleEntity.updated_by_id==User.user_id",
        post_update=True,
        doc="""The :class:`~stalker.models.auth.User` who has updated this object."""
    )

    date_created = Column(
        DateTime,
        default=datetime.datetime.now(),
        doc="""A :class:`datetime.datetime` instance showing the creation date and time of this object."""
    )

    date_updated = Column(
        DateTime,
        default=datetime.datetime.now(),
        doc="""A :class:`datetime.datetime` instance showing the update date and time of this object."""
        ,
    )

    type_id = Column(
        "type_id",
        Integer,
        ForeignKey("Types.id", use_alter=True, name="y"),
        doc="""The id of the :class:`~stalker.models.type.Type` of this entity.
        Mainly used by SQLAlchemy to create a Many-to-One relates between
        SimpleEntities and Types.
        """
    )

    type = relationship(
        "Type",
        primaryjoin="SimpleEntities.c.type_id==Types.c.id",
        post_update=True,
        doc="""The type of the object.

        It is an instance of :class:`~stalker.models.type.Type` with a proper
        :attr:`~stalker.models.type.Type.target_entity_type`.
        """
    )

    generic_data = relationship(
        'SimpleEntity',
        secondary='SimpleEntity_GenericData',
        primaryjoin='SimpleEntities.c.id==SimpleEntity_GenericData.c.simple_entity_id',
        secondaryjoin='SimpleEntity_GenericData.c.other_simple_entity_id==SimpleEntities.c.id',
        post_update=True,
        doc='''This attribute can hold any kind of data which exists in SOM.
        '''
    )

    thumbnail_id = Column(
        'thumbnail_id',
        Integer,
        ForeignKey('Links.id', use_alter=True, name='z')
    )

    thumbnail = relationship(
        'Link',
        primaryjoin='SimpleEntities.c.thumbnail_id==Links.c.id',
        post_update=True
    )

    __stalker_version__ = Column("stalker_version", String(256))

    def __init__(
            self,
            name=None,
            description="",
            type=None,
            created_by=None,
            updated_by=None,
            date_created=None,
            date_updated=None,
            thumbnail=None,
            **kwargs
    ): # pylint: disable=W0613

        # name and nice_name
        self._nice_name = ""

        self.name = name

        self.description = description
        self.created_by = created_by
        self.updated_by = updated_by
        date_created = date_created
        date_updated = date_updated
        if date_created is None:
            date_created = datetime.datetime.now()
        if date_updated is None:
            date_updated = datetime.datetime.now()

        self.date_created = date_created
        self.date_updated = date_updated
        self.type = type
        self.thumbnail = thumbnail
        self.__stalker_version__ = stalker.__version__

    # @reconstructor
    # def __init_on_load__(self):
    #     """initialized the instance variables when the instance created with
    #     SQLAlchemy
    #     """
    #     self._nice_name = None

    def __repr__(self):
        """the representation of the SimpleEntity
        """
        return "<%s (%s)>" % (self.name, self.entity_type)

    def __eq__(self, other):
        """the equality operator
        """
        return isinstance(other, SimpleEntity) and \
            self.name == other.name

    def __ne__(self, other):
        """the inequality operator
        """
        return not self.__eq__(other)

    @validates("description")
    def _validate_description(self, key, description_in):
        """validates the given description_in value
        """
        if description_in is None:
            description_in = ""

        return str(description_in)

    @validates("name")
    def _validate_name(self, key, name):
        """validates the given name_in value
        """
        if self.__auto_name__:
            if name is None or name == '':
                # generate a uuid4
                name = self.__class__.__name__ + '_' + \
                       uuid.uuid4().urn.split(':')[2]

        # it is None
        if name is None:
            raise TypeError("%s.name can not be None" %
                            self.__class__.__name__)

        if not isinstance(name, (str, unicode)):
            raise TypeError("%s.name should be an instance of string or "
                            "unicode not %s" %
                            (self.__class__.__name__,
                             name.__class__.__name__))

        name = self._format_name(str(name))

        # it is empty
        if name == "":
            raise ValueError("%s.name can not be an empty string" %
                             self.__class__.__name__)

        # also set the nice_name
        self._nice_name = self._format_nice_name(name)

        return name

    def _format_name(self, name_in):
        """formats the name_in value
        """
        # remove unnecessary characters from the string
        name_in = name_in.strip()
        #name_in = re.sub(r'([^a-zA-Z0-9\s_\-#]+)', '', name_in).strip()

        # remove all the characters which are not alphabetic from the start of
        # the string
        #name_in = re.sub(r"(^[^a-zA-Z0-9]+)", '', name_in)

        # remove multiple spaces
        name_in = re.sub(r'[\s]+', ' ', name_in)

        return name_in

    def _format_nice_name(self, nice_name_in):
        """formats the given nice name
        """
        # remove unnecessary characters from the string
        nice_name_in = str(nice_name_in).strip()
        nice_name_in = re.sub(r'([^a-zA-Z0-9\s_\-]+)', '',
                              nice_name_in).strip()

        # remove all the characters which are not alphabetic from the start of
        # the string
        nice_name_in = re.sub(r"(^[^a-zA-Z0-9]+)", '', nice_name_in)

        # remove multiple spaces
        nice_name_in = re.sub(r'[\s]+', ' ', nice_name_in)

        ## replace camel case letters
        # nice_name_in = re.sub(r"(.+?[a-z]+)([A-Z])", r"\1_\2", nice_name_in)

        # replace white spaces and dashes with under score
        nice_name_in = re.sub("([\s\-])+", r"_", nice_name_in)

        # remove multiple underscores
        nice_name_in = re.sub(r"([_]+)", r"_", nice_name_in)

        # turn it to lower case
        #nice_name_in = nice_name_in.lower()

        return nice_name_in

    @property
    def nice_name(self):
        """Nice name of this object.

        It has the same value with the name (contextually) but with a different
        format like, all the white spaces replaced by underscores ("\_"), all
        the CamelCase form will be expanded by underscore (\_) characters and
        it is always lower case.
        """
        # also set the nice_name
        #if self._nice_name is None or self._nice_name == "":
        self._nice_name = self._format_nice_name(self.name)
        return self._nice_name

    @validates("created_by")
    def _validate_created_by(self, key, created_by_in):
        """validates the given created_by_in attribute
        """
        from stalker.models.auth import User

        if created_by_in is not None:
            if not isinstance(created_by_in, User):
                raise TypeError("%s.created_by should be an instance of"
                                "stalker.models.auth.User" %
                                self.__class__.__name__)
        return created_by_in

    @validates("updated_by")
    def _validate_updated_by(self, key, updated_by_in):
        """validates the given updated_by_in attribute
        """
        from stalker.models.auth import User

        if updated_by_in is None:
            # set it to what created_by attribute has
            updated_by_in = self.created_by

        if updated_by_in is not None:
            if not isinstance(updated_by_in, User):
                raise TypeError("%s.updated_by should be an instance of"
                                "stalker.models.auth.User" %
                                self.__class__.__name__)
        return updated_by_in

    @validates("date_created")
    def _validate_date_created(self, key, date_created_in):
        """validates the given date_created_in
        """
        if date_created_in is None:
            raise TypeError("%s.date_created can not be None" %
                            self.__class__.__name__)

        if not isinstance(date_created_in, datetime.datetime):
            raise TypeError("%s.date_created should be an instance of "
                            "datetime.datetime" % self.__class__.__name__)

        return date_created_in

    @validates("date_updated")
    def _validate_date_updated(self, key, date_updated_in):
        """validates the given date_updated_in
        """
        # it is None
        if date_updated_in is None:
            raise TypeError("%s.date_updated can not be None" %
                            self.__class__.__name__)

        # it is not an instance of datetime.datetime
        if not isinstance(date_updated_in, datetime.datetime):
            raise TypeError("%s.date_updated should be an instance of "
                            "datetime.datetime" % self.__class__.__name__)

        # lower than date_created
        if date_updated_in < self.date_created:
            raise ValueError("%s.date_updated could not be set to a date "
                             "before 'date_created', try setting the "
                             "'date_created' before" %
                             self.__class__.__name__)
            # TODO: all the attribute check errors should use self.__class__.__name__ as used here
        return date_updated_in

    @validates("type")
    def _validate_type(self, key, type_in):
        """validates the given type value
        """
        from stalker.models.type import Type

        raise_error = False

        if not self.__strictly_typed__:
            if type_in is not None:
                if not isinstance(type_in, Type):
                    raise_error = True
        else:
            if not isinstance(type_in, Type):
                raise_error = True

        if raise_error:
            raise TypeError("%s.type must be an instance of "
                            "stalker.models.type.Type not %s" %
                            (self.__class__.__name__, type_in))
        return type_in

    @validates('thumbnail')
    def _validate_thumbnail(self, key, thumb):
        """validates the given thumb value
        """
        if thumb is not None:
            from stalker import Link

            if not isinstance(thumb, Link):
                raise TypeError('%s.thumbnail should be a '
                                'stalker.models.link.Link instance, not %s' %
                                (self.__class__.__name__,
                                 thumb.__class__.__name__))
        return thumb

    @property
    def tjp_id(self):
        """returns TaskJuggler compatible id
        """
        return "%s_%s" % (self.__class__.__name__, self.id)

    @property
    def to_tjp(self):
        """renders a TaskJuggler compliant string used for TaskJuggler
        integration. Needs to be overridden in inherited classes.
        """
        raise NotImplementedError('This property is not implemented in %s' %
                                  self.__class__.__name__)


class Entity(SimpleEntity):
    """Another base data class that adds tags and notes to the attributes list.

    This is the entity class which is derived from the SimpleEntity and adds
    only tags to the list of parameters.

    Two Entities considered equal if they have the same name. It doesn't matter
    if they have different tags or notes.

    :param list tags: A list of :class:`~stalker.models.tag.Tag` objects
      related to this entity. tags could be an empty list, or when omitted it
      will be set to an empty list.

    :param list notes: A list of :class:`~stalker.models.note.Note` instances.
      Can be an empty list, or when omitted it will be set to an empty list,
      when set to None it will be converted to an empty list.
    """
    __auto_name__ = True
    __tablename__ = "Entities"
    __mapper_args__ = {"polymorphic_identity": "Entity"}
    entity_id = Column("id", Integer, ForeignKey("SimpleEntities.id"),
                       primary_key=True)

    tags = relationship(
        "Tag",
        secondary="Entity_Tags",
        backref="entities",
        doc="""A list of tags attached to this object.

        It is a list of :class:`~stalker.models.tag.Tag` instances which shows
        the tags of this object"""
    )

    notes = relationship(
        "Note",
        primaryjoin="Entities.c.id==Notes.c.entity_id",
        backref="entity",
        doc="""All the :class:`~stalker.models.note.Notes`\ s attached to this entity.

        It is a list of :class:`~stalker.models.note.Note` instances or an
        empty list, setting it None will raise a TypeError.
        """
    )

    def __init__(self,
                 tags=None,
                 notes=None,
                 **kwargs):
        super(Entity, self).__init__(**kwargs)

        if tags is None:
            tags = []

        if notes is None:
            notes = []

        self.tags = tags
        self.notes = notes

    # @reconstructor
    # def __init_on_load__(self):
    #     """initialized the instance variables when the instance created with
    #     SQLAlchemy
    #     """
    #     super(Entity, self).__init_on_load__()

    @validates("notes")
    def _validate_notes(self, key, note):
        """validates the given note value
        """

        from stalker.models.note import Note

        if not isinstance(note, Note):
            raise TypeError("%s.note should be an instance of "
                            "stalker.models.note.Note not %s" %
                            (self.__class__.__name__,
                             note.__class__.__name__))

        return note

    @validates("tags")
    def _validate_tags(self, key, tag):
        """validates the given tag
        """

        from stalker.models.tag import Tag

        if not isinstance(tag, Tag):
            raise TypeError("%s.tag should be an instance of "
                            "stalker.models.tag.Tag not %s" %
                            (self.__class__.__name__,
                             tag.__class__.__name__))

        return tag

    def __eq__(self, other):
        """the equality operator
        """

        return super(Entity, self).__eq__(other) and \
               isinstance(other, Entity)

# ENTITY_TAGS
Entity_Tags = Table(
    "Entity_Tags", Base.metadata,
    Column(
        "entity_id",
        Integer,
        ForeignKey("Entities.id"),
        primary_key=True,
    ),

    Column(
        "tag_id",
        Integer,
        ForeignKey("Tags.id"),
        primary_key=True,
    )
)

# SIMPLEENTITY_GENERICDATA
SimpleEntity_GenericData = Table(
    'SimpleEntity_GenericData', Base.metadata,
    Column(
        'simple_entity_id',
        Integer,
        ForeignKey('SimpleEntities.id'),
        primary_key=True
    ),
    Column(
        'other_simple_entity_id',
        Integer,
        ForeignKey('SimpleEntities.id'),
        primary_key=True
    )
)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import subprocess
import tempfile
import datetime
import csv

import stalker
from stalker import defaults
from stalker.models.entity import Entity

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class SchedulerBase(object):
    """This is the base class for schedulers.

    All the schedulers should be derived from this class.
    """

    def __init__(self, studio=None):
        self._studio = None
        self.studio = studio

    def _validate_studio(self, studio_in):
        """validates the given studio_in value
        """
        if studio_in is not None:
            from stalker import Studio

            if not isinstance(studio_in, Studio):
                raise TypeError('%s.studio should be an instance of '
                                'stalker.models.studio.Studio, not %s' %
                                (self.__class__.__name__,
                                 studio_in.__class__.__name__))
        return studio_in

    @property
    def studio(self):
        """studio getter
        """
        return self._studio

    @studio.setter
    def studio(self, studio_in):
        """studio setter
        """
        self._studio = self._validate_studio(studio_in)

    def schedule(self):
        """the main scheduling function should be implemented in the
        derivatives
        """
        raise NotImplementedError


class TaskJugglerScheduler(SchedulerBase):
    """This is the main scheduler for Stalker right now.

    Integrates Stalker and TaskJuggler together by using TaskJugglerScheduler
    to solve the scheduling problem.

    TaskJugglerScheduler needs a :class:`~stalker.models.studio.Studio`
    instance to work with. TJS will create a .tjp file and then solve the tasks
    and restore the computed_start and computed_end dates. Combining all the
    Projects in one tjp file has a very nice side effect, projects using the
    same resources will respect their allocations to the resource.
    """

    def __init__(self, studio=None):
        super(TaskJugglerScheduler, self).__init__(studio)

        self.tjp_content = ''

        self.temp_file_full_path = None
        self.temp_file_path = None
        self.temp_file_name = None

        self.tjp_file_full_path = None
        self.tjp_file = None

        self.csv_file_full_path = None
        self.csv_file = None

    def _create_tjp_file(self):
        """creates the tjp file
        """
        self.temp_file_full_path = tempfile.mktemp(prefix='Stalker_')
        self.temp_file_name = os.path.basename(self.temp_file_full_path)
        self.tjp_file_full_path = self.temp_file_full_path + ".tjp"
        self.csv_file_full_path = self.temp_file_full_path + ".csv"

    def _create_tjp_file_content(self):
        """creates the tjp file content
        """
        from jinja2 import Template

        template = Template(defaults.tjp_main_template)

        self.tjp_content = template.render({
            'stalker': stalker,
            'studio': self.studio,
            'csv_file_name': self.temp_file_name,
            'csv_file_full_path': self.temp_file_full_path
        })

    def _fill_tjp_file(self):
        """fills the tjp file with content
        """
        with open(self.tjp_file_full_path, 'w+') as self.tjp_file:
            self.tjp_file.write(self.tjp_content)

    def _delete_tjp_file(self):
        """deletes the temp tjp file
        """
        try:
            os.remove(self.tjp_file_full_path)
        except OSError:
            pass

    def _delete_csv_file(self):
        """deletes the temp csv file
        """
        try:
            os.remove(self.csv_file_full_path)
        except OSError:
            pass

    def _clean_up(self):
        """removes the temp files
        """
        self._delete_tjp_file()
        self._delete_csv_file()

    def _parse_csv_file(self):
        """parses back the csv file and fills the tasks with computes_start and
        computed end values
        """
        logger.debug('csv_file_full_path : %s' % self.csv_file_full_path)

        with open(self.csv_file_full_path, 'r') as self.csv_file:
            csv_content = csv.reader(self.csv_file, delimiter=';')
            lines = [line for line in csv_content]
            lines.pop(0)
            for data in lines:
                id_line = data[0]
                entity_id = int(id_line.split('.')[-1].split('_')[-1])
                logger.debug('updating entity with id : %s' % entity_id)
                entity = Entity.query.filter(Entity.id == entity_id).first()
                logger.debug('updating : %s' % entity)
                if entity:
                    start_date = datetime.datetime.strptime(
                        data[1], "%Y-%m-%d-%H:%M"
                    )
                    end_date = datetime.datetime.strptime(
                        data[2],
                        "%Y-%m-%d-%H:%M"
                    )
                    logger.debug('   start : %s' % start_date)
                    logger.debug('     end : %s' % end_date)

                    entity.computed_start = start_date
                    entity.computed_end = end_date
                    logger.debug('updated : %s' % entity)
        logger.debug('completed parsing csv file')

    def schedule(self):
        """does the scheduling
        """
        # check the studio attribute
        from stalker import Studio

        if not isinstance(self.studio, Studio):
            raise TypeError('%s.studio should be an instance of '
                            'stalker.models.studio.Studio, not %s' %
                            (self.__class__.__name__,
                             self.studio.__class__.__name__))

        # ********************************************************************
        # Adjust Studio.start and Studio.end
        # get the active projects
        # get root tasks
        # calculate the earliest start and the latest end
        self.studio._start = datetime.datetime(2013, 1, 1) # TODO: remove the hardcoded datetime
        self.studio._end = datetime.datetime(2016, 1, 1) # TODO: remove the hardcoded datetime
        #for project in self.studio.active_projects:
        #    for root_task in project.root_tasks:
        #        if root_task.start < self.studio.start:
        #            self.studio.start = root_task.start
        #        if root_task.end > self.studio.end:
        #            self.studio.end = root_task.end

        # now for safety multiply the duration by 2
        #self.studio.end = (self.studio.end - self.studio.start) * 5 + \
        #                  self.studio.start

        # create a tjp file
        self._create_tjp_file()

        # create tjp file content
        self._create_tjp_file_content()

        # fill it with data
        self._fill_tjp_file()

        logger.debug('tjp_file_full_path: %s' % self.tjp_file_full_path)

        # pass it to tj3
        proc = subprocess.Popen(
            [defaults.tj_command,
             self.tjp_file_full_path],
            stderr=subprocess.PIPE
        )
        # wait it to complete
        proc.wait()

        if proc.returncode:
            raise RuntimeError(proc.stderr.readlines())

        logger.debug('tj3 return code: %s' % proc.returncode)
        logger.debug('tj3 output: %s' % proc.stderr.readlines())
        # read back the csv file
        self._parse_csv_file()

        # remove the tjp file
        #self._clean_up()





# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest2
from stalker import Note

class NoteTester(unittest2.TestCase):
    """tests  the Note class
    """
    def setUp(self):
        """setup the test
        """
        self.kwargs = {
            "name": "Note to something",
            "description": "this is a simple note",
            "content": "this is a note content",
            }

        # create a Note object
        self.test_note = Note(**self.kwargs)
    
    def test___auto_name__class_attribute_is_set_to_True(self):
        """testing if the __auto_name__ class attribute is set to True for
        Note class
        """
        self.assertTrue(Note.__auto_name__)
    
    def test_content_argument_is_missing(self):
        """testing if nothing is going to happen when no content argument is
        given
        """
        self.kwargs.pop("content")
        new_note = Note(**self.kwargs)

    def test_content_argument_is_set_to_None(self):
        """testing if nothing is going to happen when content argument is given
        as None
        """
        self.kwargs["content"] = None
        new_note = Note(**self.kwargs)

    def test_content_attribute_is_set_to_None(self):
        """testing if nothing is going to happen when content attribute is set
        to None
        """
        # nothing should happen
        self.test_note.content = None

    def test_content_argument_is_set_to_empty_string(self):
        """testing if nothing is going to happen when content argument is given
        as an empty string
        """
        self.kwargs["content"] = ""
        new_note = Note(**self.kwargs)

    def test_content_attribute_is_set_to_empty_string(self):
        """testing if nothing is goin to happen when content argument is set to
        an empty string
        """
        # nothing should happen
        self.test_note.content = ""

    def test_content_argument_is_set_to_something_other_than_a_string(self):
        """testing if a TypeEror will be raised when trying to set the content
        argument to something other than a string or unicode
        """
        test_values = [1, 1.24, ["content"], {"a": "Content"}]

        for test_value in test_values:
            self.kwargs["content"] = test_value
            self.assertRaises(TypeError, Note, **self.kwargs)

    def test_content_attribute_is_set_to_something_other_than_a_string(self):
        """testing if a TypeError will be raised when trying to set the
        content attribute to something other than a string or unicode
        """
        test_values = [1, 1.24, ["content"], {"a": "Content"}]

        for test_value in test_values:
            self.assertRaises(
                TypeError,
                setattr,
                self.test_note,
                "content",
                test_value
            )

    def test_content_attribute_is_working_properly(self):
        """testing if the content attribute is working properly
        """
        new_content = "This is my new content for the note, and I expect it to\
        work fine when I assign it to a Note object"

        self.test_note.content = new_content

        self.assertEqual(self.test_note.content, new_content)

    def test_equality_operator(self):
        """testing equality operator
        """
        note1 = Note(**self.kwargs)
        note2 = Note(**self.kwargs)

        self.kwargs["content"] = "this is a different content"
        note3 = Note(**self.kwargs)

        self.assertTrue(note1 == note2)
        self.assertFalse(note1 == note3)

    def test_inequality_operator(self):
        """testing inequality operator
        """
        note1 = Note(**self.kwargs)
        note2 = Note(**self.kwargs)

        self.kwargs["content"] = "this is a different content"
        note3 = Note(**self.kwargs)

        self.assertFalse(note1 != note2)
        self.assertTrue(note1 != note3)

    def test_plural_class_name(self):
        """testing the plural name of Note class
        """
        self.assertTrue(self.test_note.plural_class_name, "Notes")



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import copy
import logging
import datetime

from sqlalchemy import Column, Integer, ForeignKey, Table
from sqlalchemy.orm import validates, reconstructor, relationship

from stalker import defaults, log
from stalker.models.entity import SimpleEntity, Entity
from stalker.models.mixins import ScheduleMixin, WorkingHoursMixin
from stalker.models.schedulers import SchedulerBase

logger = logging.getLogger(__name__)
logger.setLevel(log.logging_level)


class Studio(Entity, ScheduleMixin, WorkingHoursMixin):
    """Manage all the studio information at once.

    With Stalker you can manage all you Studio data by using this class. Studio
    knows all the projects, all the departments, all the users and every thing
    about the studio. But the most important part of the Studio is that it can
    schedule all the Projects by using TaskJuggler.

    Studio class is kind of the database itself::

      studio = Studio()

      # simple data
      studio.projects
      studio.active_projects
      studio.inactive_projects
      studio.departments
      studio.users

      # project management
      studio.to_tjp          # a tjp representation of the studio with all
                             # its projects, departments and resources etc.

      studio.schedule() # schedules all the active projects at once

    **Working Hours**

    In Stalker, Studio class also manages the working hours of the studio.
    Allowing project tasks to be scheduled to be scheduled in those hours.

    **Vacations**

    You can define

    :param int daily_working_hours: An integer specifying the daily working
      hours for the studio. It is another critical value attribute which
      TaskJuggler uses mainly converting working day values to working hours
      (1d = 10h kind of thing).

    :param now: The now attribute overrides the TaskJugglers ``now`` attribute
      allowing the user to schedule the projects as if the scheduling is done
      on that date. The default value is the rounded value of
      datetime.datetime.now().

    :type now: datetime.datetime
    """
    __auto_name__ = False
    __tablename__ = 'Studios'
    __mapper_args__ = {'polymorphic_identity': 'Studio'}

    studio_id = Column(
        'id',
        Integer,
        ForeignKey('Entities.id'),
        primary_key=True,
    )

    daily_working_hours = Column(Integer, default=8)

    def __init__(self,
                 daily_working_hours=None,
                 now=None,
                 **kwargs):
        super(Studio, self).__init__(**kwargs)
        ScheduleMixin.__init__(self, **kwargs)
        WorkingHoursMixin.__init__(self, **kwargs)
        # TODO: daily_working_hours should be in WorkingHours not in Studio
        self.daily_working_hours = daily_working_hours
        self._now = None
        self.now = self._validate_now(now)
        self._scheduler = None

    # @reconstructor
    # def __init_on_load__(self):
    #     self._now = None
    #     super(Studio, self).__init_on_load__()

    @validates('daily_working_hours')
    def _validate_daily_working_hours(self, key, dwh):
        """validates the given daily working hours value
        """
        if dwh is None:
            dwh = defaults.daily_working_hours
        else:
            if not isinstance(dwh, int):
                raise TypeError('%s.daily_working_hours should be an integer, '
                                'not %s' %
                                (self.__class__.__name__,
                                 dwh.__class__.__name__))
        return dwh

    def _validate_now(self, now_in):
        """validates the given now_in value
        """
        if now_in is None:
            now_in = datetime.datetime.now()

        if not isinstance(now_in, datetime.datetime):
            raise TypeError('%s.now attribute should be an instance of '
                            'datetime.datetime, not %s' %
                            (self.__class__.__name__,
                             now_in.__class__.__name__))

        return self.round_time(now_in)

    @property
    def now(self):
        """now getter
        """
        try:
            if self._now is None:
                self._now = self.round_time(datetime.datetime.now())
        except AttributeError:
            setattr(self, '_now', self.round_time(datetime.datetime.now()))
        return self._now

    @now.setter
    def now(self, now_in):
        """now setter
        """
        self._now = self._validate_now(now_in)

    def _validate_scheduler(self, scheduler_in):
        """validates the given scheduler_in value
        """
        if scheduler_in is not None:
            if not isinstance(scheduler_in, SchedulerBase):
                raise TypeError('%s.scheduler should be an instance of '
                                'stalker.models.scheduler.SchedulerBase, not '
                                '%s' % (self.__class__.__name__,
                                        scheduler_in.__class__.__name__))
        return scheduler_in

    @property
    def scheduler(self):
        """scheduler getter
        """
        return self._scheduler

    @scheduler.setter
    def scheduler(self, scheduler_in):
        """the scheduler setter
        """
        self._scheduler = self._validate_scheduler(scheduler_in)

    @property
    def to_tjp(self):
        """converts the studio to a tjp representation
        """
        from jinja2 import Template

        temp = Template(defaults.tjp_studio_template)
        return temp.render({
            'studio': self,
            'datetime': datetime,
            'now': self.round_time(self.now).strftime('%Y-%m-%d-%H:%M')
        })

    @property
    def projects(self):
        """returns all the projects in the studio
        """
        from stalker import Project

        return Project.query.all()

    @property
    def active_projects(self):
        """returns all the active projects in the studio
        """
        from stalker import Project

        return Project.query.filter(Project.active == True).all()

    @property
    def inactive_projects(self):
        """return all the inactive projects in the studio
        """
        from stalker import Project

        return Project.query.filter(Project.active == False).all()

    @property
    def departments(self):
        """returns all the departments in the studio
        """
        from stalker import Department

        return Department.query.all()

    @property
    def users(self):
        """returns all the users in the studio
        """
        from stalker import User

        return User.query.all()

    def schedule(self):
        """Schedules all the active projects in the studio. Needs a Scheduler,
        so before calling it set a scheduler by using the :attr:`.scheduler`
        attribute.
        """
        # check the scheduler first
        if self.scheduler is None or \
                not isinstance(self.scheduler, SchedulerBase):
            raise RuntimeError('There is no scheduler for this %s, please '
                               'assign a scheduler to the %s.scheduler '
                               'attribute, before calling %s.schedule()' %
                               (self.__class__.__name__,
                                self.__class__.__name__,
                                self.__class__.__name__))

        # run the scheduler
        self.scheduler.studio = self
        self.scheduler.schedule()

    @property
    def weekly_working_hours(self):
        """returns the WorkingHours.weekly_working_hours
        """
        return self.working_hours.weekly_working_hours

    @property
    def weekly_working_days(self):
        """returns the WorkingHours.weekly_working_hours
        """
        return self.working_hours.weekly_working_days

    @property
    def yearly_working_days(self):
        """returns the yearly working days
        """
        return self.working_hours.yearly_working_days


class WorkingHours(object):
    """A helper class to manage Studio working hours.

    Working hours is a data class to store the weekly working hours pattern of
    the studio. 

    The data stored as a dictionary with the short day names are used as the
    key and the value is a list of two integers showing the working hours
    interval as the minutes after midnight. This is done in that way to ease
    the data transfer to TaskJuggler. The default value is defined in
    :class:`stalker.config.Config` ::

      wh = WorkingHours()
      wh.working_hours = {
          'mon': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'tue': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'wed': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'thu': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'fri': [[540, 720], [820, 1080]], # 9:00 - 12:00, 13:00 - 18:00
          'sat': [], # saturday off
          'sun': [], # sunday off
      }

    The default value is 9:00 - 18:00 from Monday to Friday and Saturday and
    Sunday are off.

    The working hours can be updated by the user supplied dictionary. If the
    user supplied dictionary doesn't have all the days then the default values
    will be used for those days.

    It is possible to use day index and day short names as a key value to reach
    the data::

      from stalker import config
      defaults = config.Config()

      wh = WorkingHours()

      # this is same by doing wh.working_hours['sun']
      assert wh['sun'] == defaults.working_hours['sun']

      # you can reach the data using the weekday number as index
      assert wh[0] == defaults.working_hours['mon']

      # working hours of sunday if defaults are used or any other day defined
      # by the stalker.config.Config.day_order
      assert wh[0] == defaults.working_hours[defaults.day_order[0]]

    :param working_hours: The dictionary that shows the working hours. The keys
      of the dictionary should be one of ['mon', 'tue', 'wed', 'thu', 'fri',
      'sat', 'sun']. And the values should be a list of two integers like
      [[int, int], [int, int], ...] format, showing the minutes after midnight.
      For missing days the default value will be used. If skipped the default
      value is going to be used.
    """

    def __init__(self, working_hours=None, **kwargs):
        if working_hours is None:
            working_hours = defaults.working_hours
        self._wh = None
        self.working_hours = self._validate_working_hours(working_hours)

    def __eq__(self, other):
        """equality test
        """
        return isinstance(other, WorkingHours) and \
               other.working_hours == self.working_hours

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._wh[defaults.day_order[item]]
        elif isinstance(item, str):
            return self._wh[item]

    def __setitem__(self, key, value):
        self._validate_wh_value(value)
        if isinstance(key, int):
            self._wh[defaults.day_order[key]] = value
        elif isinstance(key, str):
            # check if key is in
            if key not in defaults.day_order:
                raise KeyError('%s accepts only %s as key, not %s' %
                               (self.__class__.__name__, defaults.day_order,
                                key))
            self._wh[key] = value

    def _validate_working_hours(self, wh_in):
        """validates the given working hours
        """
        if not isinstance(wh_in, dict):
            raise TypeError('%s.working_hours should be a dictionary, not %s' %
                            (self.__class__.__name__,
                             wh_in.__class__.__name__))

        for key in wh_in.keys():
            if not isinstance(wh_in[key], list):
                raise TypeError('%s.working_hours should be a dictionary with '
                                'keys "mon, tue, wed, thu, fri, sat, sun" '
                                'and the values should a list of lists of '
                                'two integers like [[540, 720], [800, 1080]], '
                                'not %s' % (self.__class__.__name__,
                                            wh_in[key].__class__.__name__))

            # validate item values
            self._validate_wh_value(wh_in[key])

        # update the default values with the supplied working_hour dictionary
        # copy the defaults
        wh_def = copy.copy(defaults.working_hours)
        # update them
        wh_def.update(wh_in)

        return wh_def

    @property
    def working_hours(self):
        """the getter of _wh
        """
        return self._wh

    @working_hours.setter
    def working_hours(self, wh_in):
        """the setter of _wh
        """
        self._wh = self._validate_working_hours(wh_in)

    def is_working_hour(self, check_for_date):
        """checks if the given datetime is in working hours

        :param datetime.datetime check_for_date: The time to check if it is a
          working hour
        """
        weekday_nr = check_for_date.weekday()
        hour = check_for_date.hour
        minute = check_for_date.minute

        time_from_midnight = hour * 60 + minute

        # check if the hour is inside the working hour ranges
        logger.debug('checking for: %s' % time_from_midnight)
        logger.debug('self[weekday_nr]: %s' % self[weekday_nr])
        for working_hour_groups in self[weekday_nr]:
            start = working_hour_groups[0]
            end = working_hour_groups[1]
            logger.debug('start       : %s' % start)
            logger.debug('end         : %s' % end)
            if start <= time_from_midnight < end:
                return True

        return False

    def _validate_wh_value(self, value):
        """validates the working hour value
        """
        err = '%s.working_hours value should be a list of lists of two ' \
              'integers between and the range of integers should be 0-1440, ' \
              'not %s'

        if not isinstance(value, list):
            raise TypeError(err % (self.__class__.__name__,
                                   value.__class__.__name__))

        for i in value:
            if not isinstance(i, list):
                raise TypeError(err % (self.__class__.__name__,
                                       i.__class__.__name__))

            # check list length
            if len(i) != 2:
                raise RuntimeError(err % (self.__class__.__name__, value))

            # check type
            if not isinstance(i[0], int) or not isinstance(i[1], int):
                raise TypeError(err % (self.__class__.__name__, value))

            # check range
            if i[0] < 0 or i[0] > 1440 or i[1] < 0 or i[1] > 1440:
                raise ValueError(err % (self.__class__.__name__, value))

        return value

    @property
    def to_tjp(self):
        """returns TaskJuggler representation of this object
        """
        # render the template
        from jinja2 import Template

        template = Template(defaults.tjp_working_hours_template)
        return template.render({'workinghours': self})

    @property
    def weekly_working_hours(self):
        """returns the total working hours in a week
        """
        weekly_working_hours = 0
        for i in range(0, 7):
            for start, end in self[i]:
                weekly_working_hours += (end - start)
        return weekly_working_hours / 60.0

    @property
    def weekly_working_days(self):
        """returns the weekly working days by looking at the working hours
        settings
        """
        wwd = 0
        for i in range(0, 7):
            if len(self[i]):
                wwd += 1
        return wwd

    @property
    def yearly_working_days(self):
        """returns the total working days in a year
        """
        return self.weekly_working_days * 52.1428


class Vacation(SimpleEntity, ScheduleMixin):
    """Vacation is the way to manage the User vacations.

    :param user: The user of this vacation. Should be an instance of
      :class:`~stalker.models.auth.User` and can not be skipped or can not be
      None.

    :param start: The start datetime of the vacation. Is is an
      datetime.datetime instance. When skipped it will be set to the rounded
      value of 

    :param end: The end datetime of the vacation. It is an datetime.datetime
      instance.
    """
    __auto_name__ = True
    __tablename__ = 'Vacations'
    __mapper_args__ = {'polymorphic_identity': 'Vacation'}

    __strictly_typed__ = True

    vacation_id = Column("id", Integer, ForeignKey("SimpleEntities.id"),
                         primary_key=True)

    user_id = Column('user_id', Integer, ForeignKey('Users.id'),
                     nullable=False)

    user = relationship(
        'User',
        primaryjoin='Vacations.c.user_id==Users.c.id',
        back_populates='vacations',
        doc="""The User of this Vacation.

        Accepts:class:`~stalker.models.auth.User` instance.
        """
    )

    def __init__(self, user=None, start=None, end=None, **kwargs):
        kwargs['start'] = start
        kwargs['end'] = end
        super(Vacation, self).__init__(**kwargs)
        ScheduleMixin.__init__(self, **kwargs)
        self.user = user

    @validates('user')
    def _validate_user(self, key, user):
        """validates the given user instance
        """
        from stalker import User

        if not isinstance(user, User):
            raise TypeError('%s.user should be an instance of '
                            'stalker.models.auth.User, not %s' %
                            (self.__class__.__name__, user.__class__.__name__))
        return user

    @property
    def to_tjp(self):
        """overridden to_tjp method
        """
        from jinja2 import Template

        template = Template(defaults.tjp_vacation_template)
        return template.render({'vacation': self})



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README')).read()
CHANGES = open(os.path.join(here, 'CHANGELOG')).read()

requires = [
    'sqlalchemy>=0.8',
    'alembic',
    'jinja2',
    'unittest2',
    #'sphinx==1.1.3',
]

setup(name='stalker',
      version='0.2.0.rc4',
      description='A Production Asset Management (ProdAM) System',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Database",
        "Topic :: Software Development",
        "Topic :: Utilities",
        "Topic :: Office/Business :: Scheduling",
      ],
      author='Erkan Ozgur Yilmaz',
      author_email='eoyilmaz@gmail.com',
      url='http://code.google.com/p/stalker/',
      keywords=['production', 'asset', 'management', 'vfx', 'animation',
                'maya', 'houdini', 'nuke', 'fusion', 'softimage', 'blender',
                'vue'],
      packages=find_packages(),
      include_package_data=True,
      zip_safe=True,
      test_suite='stalker',
      install_requires=requires,
)




# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest2
from stalker import (Asset, Entity, Project, Repository, Sequence, Status,
                     StatusList, Task, Type, Link, Shot)

class AssetTester(unittest2.TestCase):
    """tests Asset class    """    

    def setUp(self):
        """setup the test
        """
        # statuses
        self.test_data_status_complete = Status(
            name="Complete",
            code="CMPLT",
        )
        
        self.test_data_status_wip = Status(
            name="Work In Progress",
            code="WIP",
        )
        
        # status lists
        self.test_data_project_status_list = StatusList(
            name="Project Status List",
            target_entity_type=Project,
            statuses=[
                self.test_data_status_complete,
                self.test_data_status_wip
            ]
        )

        self.test_data_task_status_list = StatusList(
            name="Task Status List",
            target_entity_type=Task,
            statuses=[
                self.test_data_status_complete,
                self.test_data_status_wip
            ]
        )

        self.test_data_asset_status_list = StatusList(
            name="Asset Status List",
            target_entity_type=Asset,
            statuses=[
                self.test_data_status_complete,
                self.test_data_status_wip
            ]
        )

        self.test_data_shot_status_list = StatusList(
            name="Shot Status List",
            target_entity_type=Shot,
            statuses=[
                self.test_data_status_complete,
                self.test_data_status_wip
            ]
        )

        self.test_data_sequence_status_list = StatusList(
            name="Sequence Status List",
            target_entity_type=Sequence,
            statuses=[
                self.test_data_status_complete,
                self.test_data_status_wip
            ]
        )

        # types
        self.test_data_commmercial_project = Type(
            name="Commercial Project",
            code='commproj',
            target_entity_type=Project,
        )

        self.test_data_asset_type1 = Type(
            name="Character",
            code='char',
            target_entity_type=Asset
        )

        self.test_data_asset_type2 = Type(
            name="Environment",
            code='env',
            target_entity_type=Asset
        )

        self.test_data_repository_type = Type(
            name="Test Repository Type",
            code='testrepo',
            target_entity_type=Repository,
            )

        # repository
        self.test_data_repository = Repository(
            name="Test Repository",
            type=self.test_data_repository_type,
        )

        # project
        self.test_data_project1 = Project(
            name="Test Project1",
            code='tp1',
            type=self.test_data_commmercial_project,
            status_list=self.test_data_project_status_list,
            repository=self.test_data_repository,
        )

        # sequence
        self.test_data_sequence = Sequence(
            name="Test Sequence",
            code='tseq',
            project=self.test_data_project1,
            status_list=self.test_data_sequence_status_list,
        )

        # shots
        self.test_data_shot1 = Shot(
            code="TestSH001",
            status_list=self.test_data_shot_status_list,
            project=self.test_data_project1,
            sequences=[self.test_data_sequence],
        )
        
        self.test_data_shot2 = Shot(
            code="TestSH002",
            status_list=self.test_data_shot_status_list,
            project=self.test_data_project1,
            sequences=[self.test_data_sequence],
        )

        self.test_data_shot3 = Shot(
            code="TestSH003",
            status_list=self.test_data_shot_status_list,
            project=self.test_data_project1,
            sequences=[self.test_data_sequence],
        )
        
        self.test_data_shot4 = Shot(
            code="TestSH004",
            status_list=self.test_data_shot_status_list,
            project=self.test_data_project1,
            sequences=[self.test_data_sequence],
        )
        
        self.kwargs = {
            "name": "Test Asset",
            'code': 'ta',
            "description": "This is a test Asset object",
            "project": self.test_data_project1,
            "type": self.test_data_asset_type1,
            "status": 0,
            "status_list": self.test_data_asset_status_list,
        }
        
        self.test_data_test_asset = Asset(**self.kwargs)
        
        # tasks
        self.test_data_task1 = Task(
            name="Task1",
            parent=self.test_data_test_asset,
            status_list=self.test_data_task_status_list
        )

        self.test_data_task2 = Task(
            name="Task2",
            parent=self.test_data_test_asset,
            status_list=self.test_data_task_status_list
        )

        self.test_data_task3 = Task(
            name="Task3",
            parent=self.test_data_test_asset,
            status_list=self.test_data_task_status_list
        )
    
    def test___auto_name__class_attribute_is_set_to_False(self):
        """testing if the __auto_name__ class attribute is set to False for
        Asset class
        """
        self.assertFalse(Asset.__auto_name__)
    
    def test_equality(self):
        """testing equality of two Asset objects
        """

        new_asset1 = Asset(**self.kwargs)
        new_asset2 = Asset(**self.kwargs)

        new_entity1 = Entity(**self.kwargs)

        self.kwargs["type"] = self.test_data_asset_type2
        new_asset3 = Asset(**self.kwargs)

        self.kwargs["name"] = "another name"
        new_asset4 = Asset(**self.kwargs)

        self.assertTrue(new_asset1 == new_asset2)
        self.assertFalse(new_asset1 == new_asset3)
        self.assertFalse(new_asset1 == new_asset4)
        self.assertFalse(new_asset3 == new_asset4)
        self.assertFalse(new_asset1 == new_entity1)

    def test_inequality(self):
        """testing inequality of two Asset objects
        """

        new_asset1 = Asset(**self.kwargs)
        new_asset2 = Asset(**self.kwargs)

        new_entity1 = Entity(**self.kwargs)

        self.kwargs["type"] = self.test_data_asset_type2
        new_asset3 = Asset(**self.kwargs)

        self.kwargs["name"] = "another name"
        new_asset4 = Asset(**self.kwargs)

        self.assertFalse(new_asset1 != new_asset2)
        self.assertTrue(new_asset1 != new_asset3)
        self.assertTrue(new_asset1 != new_asset4)
        self.assertTrue(new_asset3 != new_asset4)
        self.assertTrue(new_asset1 != new_entity1)
    
    def test_ReferenceMixin_initialization(self):
        """testing if the ReferenceMixin part is initialized correctly
        """

        link_type_1 = Type(
            name="Image",
            code='image',
            target_entity_type="Link"
        )
        
        link1 = Link(
            name="Artwork 1",
            full_path="/mnt/M/JOBs/TEST_PROJECT",
            filename="a.jpg",
            type=link_type_1
        )
        
        link2 = Link(
            name="Artwork 2",
            full_path="/mnt/M/JOBs/TEST_PROJECT",
            filename="b.jbg",
            type=link_type_1
        )
        
        references = [link1, link2]
        
        self.kwargs["code"] = "SH12314"
        self.kwargs["references"] = references
        
        new_asset = Asset(**self.kwargs)
        
        self.assertEqual(new_asset.references, references)
    
    def test_StatusMixin_initialization(self):
        """testing if the StatusMixin part is initialized correctly
        """

        status1 = Status(name="On Hold", code="OH")
        status2 = Status(name="Complete", code="CMPLT")

        status_list = StatusList(name="Project Statuses",
                                 statuses=[status1, status2],
                                 target_entity_type=Asset)
        
        self.kwargs["code"] = "SH12314"
        self.kwargs["status"] = 0
        self.kwargs["status_list"] = status_list

        new_asset = Asset(**self.kwargs)

        self.assertEqual(new_asset.status_list, status_list)

    def test_TaskMixin_initialization(self):
        """testing if the TaskMixin part is initialized correctly
        """

        status1 = Status(name="On Hold", code="OH")

        task_status_list = StatusList(
            name="Task Statuses",
            statuses=[status1],
            target_entity_type=Task
        )

        project_status_list = StatusList(
            name="Project Statuses",
            statuses=[status1],
            target_entity_type=Project,
            )

        commercial_project_type = Type(
            name="Commercial",
            code='comm',
            target_entity_type=Project,
            )

        new_project = Project(
            name="Commercial",
            code='COM',
            type=commercial_project_type,
            status_list=project_status_list,
            repository=self.test_data_repository,
        )

        character_asset_type = Type(
            name="Character",
            code='char',
            target_entity_type=Asset
        )

        asset_status_list = StatusList(
            name="Asset Status List",
            statuses=[status1],
            target_entity_type=Asset
        )

        new_asset = Asset(
            name="test asset",
            type=character_asset_type,
            code="tstasset",
            status_list=asset_status_list,
            project=new_project,
            )

        task1 = Task(
            name="Modeling",
            status_list=task_status_list,
            parent=new_asset
        )

        task2 = Task(
            name="Lighting",
            status_list=task_status_list,
            parent=new_asset
        )

        tasks = [task1, task2]

        self.assertItemsEqual(new_asset.tasks, tasks)


        # TODO: this test should be in tests.db
        # because the property it is testing is using DBSession.query
        #
        #def test_asset_appends_itself_to_the_assets_list_of_project_instance(self):
        #"""testing if the created Asset instance will append itself to the
        #assets list of the given Project instance.
        #"""

        #status1 = Status(name="On Hold", code="OH")

        #task_status_list = StatusList(
        #name="Task Statuses",
        #statuses=[status1],
        #target_entity_type=Task
        #)

        #project_status_list = StatusList(
        #name="Project Statuses",
        #statuses=[status1],
        #target_entity_type=Project
        #)

        #commercial_project_type = Type(
        #name="Commercial",
        #code='comm',
        #target_entity_type=Project,
        #)

        #new_project = Project(
        #name="Commercial",
        #type=commercial_project_type,
        #status_list=project_status_list,
        #repository=self.test_data_repository,
        #)

        #character_asset_type = Type(
        #name="Character",
        #code='char',
        #target_entity_type=Asset
        #)

        #asset_status_list = StatusList(
        #name="Asset Status List",
        #statuses=[status1],
        #target_entity_type=Asset
        #)

        #new_asset = Asset(
        #   name="test asset",
        #   type=character_asset_type,
        #   code="tstasset",
        #   status_list=asset_status_list,
        #   project=new_project,
        #)

        #task1 = Task(
        #   name="Modeling",
        #   status_list=task_status_list,
        #   parent=new_asset
        #)

        #task2 = Task(
        #   name="Lighting",
        #   status_list=task_status_list,
        #   parent=new_asset
        #)

        #self.assertIn(new_asset, new_project.assets)

    def test_plural_class_name(self):
        """testing the default plural name of the Asset class
        """
        self.assertEqual(self.test_data_test_asset.plural_class_name, "Assets")

    def test___strictly_typed___is_True(self):
        """testing if the __strictly_typed__ class attribute is True
        """

        self.assertEqual(Asset.__strictly_typed__, True)

#    def test_shots_argument_is_skipped(self):
#        """testing if the shots attribute will be an empty list when the shots
#        argument is skipped
#        """
#        self.kwargs.pop("shots")
#        new_asset = Asset(**self.kwargs)
#        self.assertEqual(new_asset.shots, [])
#
#    def test_shots_argument_is_None(self):
#        """testing if the shots attribute will be an empty list when the shots
#        argument is None
#        """
#        self.kwargs["shots"] = None
#        new_asset = Asset(**self.kwargs)
#        self.assertEqual(new_asset.shots, [])
#
#    def test_shots_attribute_is_None(self):
#        """testing if a TypeError will be raised when the shots attribute is
#        set to None
#        """
#        self.assertRaises(TypeError, setattr, self.test_data_test_asset,
#                          "shots", None)
#
#    def test_shots_argument_is_not_a_list(self):
#        """testing if a TypeError will be raised when the shots argument is not
#        a list
#        """
#        self.kwargs["shots"] = "1"
#        self.assertRaises(TypeError, Asset, **self.kwargs)
#
#    def test_shots_attribute_is_not_a_list(self):
#        """testing if a TypeError will be raised when the shots attribute is
#        set to a value other than a list
#        """
#        self.assertRaises(TypeError, setattr, self.test_data_test_asset,
#                          "shots", "1")
#
#    def test_shots_argument_is_not_a_list_of_Shots(self):
#        """testing if a TypeError will be raised when the shots argument is not
#        a list of Shot instances
#        """
#        self.kwargs["shots"] = [1, 1.2, "a shot"]
#        self.assertRaises(TypeError, Asset, **self.kwargs)
#
#    def test_shots_attribute_is_not_a_list_of_Shots(self):
#        """testing if a TypeError will be raised when the shots attribute is
#        set to a value other than a list of Shot instances
#        """
#        self.assertRaises(TypeError, setattr, self.test_data_test_asset,
#                          "shots", [1, 1.2, "a shot"])
#
#    def test_shots_argument_is_a_list_of_Shot_instances(self):
#        """testing if the assets attribute of the Shot instances will be
#        updated and have the current asset in their assets list when the shots
#        argument is a list of Shot instances
#        """
#
#        self.kwargs["shots"] = [self.test_data_shot1,
#                                self.test_data_shot2]
#
#        new_asset = Asset(**self.kwargs)
#
#        self.assertIn(new_asset, self.test_data_shot1.assets)
#        self.assertIn(new_asset, self.test_data_shot2.assets)
#
#    def test_shots_attribute_is_a_list_of_Shot_instances(self):
#        """testing if the assets attribute of the Shot instances will be
#        updated and have the current asset in their assets list when the shots
#        attribute is a list of Shot instances
#        """
#
#        self.kwargs["name"] = "New Test Asset"
#        self.kwargs["shots"] = [self.test_data_shot1,
#                                self.test_data_shot2]
#
#        #print "creating new asset"
#        new_asset = Asset(**self.kwargs)
#
#        #print "appending new shots"
#        new_asset.shots = [self.test_data_shot3,
#                           self.test_data_shot4]
#
#        self.assertIn(new_asset, self.test_data_shot3.assets)
#        self.assertIn(new_asset, self.test_data_shot4.assets)
#
#        self.assertNotIn(new_asset, self.test_data_shot1.assets)
#        self.assertNotIn(new_asset, self.test_data_shot2.assets)
#
#    def test_shots_attribute_will_update_the_backreference_value_assets_in_Shot_instances(self):
#        """testing if the shots attribute will update the backreference
#        attribute in Shot instances
#        """
#
#        self.kwargs["name"] = "New Test Asset"
#        self.kwargs["shots"] = [self.test_data_shot1, self.test_data_shot2]
#
#        new_asset = Asset(**self.kwargs)
#
#        # append
#        new_asset.shots.append(self.test_data_shot3)
#        self.assertIn(new_asset, self.test_data_shot3.assets)
#
#        # extend
#        new_asset.shots.extend([self.test_data_shot4])
#        self.assertIn(new_asset, self.test_data_shot4.assets)
#
#        # remove
#        new_asset.shots.remove(self.test_data_shot1)
#        self.assertNotIn(new_asset, self.test_data_shot1.assets)
#
#        # pop
#        new_asset.shots.pop(0)
#        self.assertNotIn(new_asset, self.test_data_shot2.assets)
#
#        # pop again
#        new_asset.shots.pop()
#        self.assertNotIn(new_asset, self.test_data_shot4.assets)
#    



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
import os
import pickle
import re
import base64
import datetime

from sqlalchemy import (Table, Column, Integer, ForeignKey, String, DateTime,
                        Enum)
from sqlalchemy.orm import relationship, synonym, reconstructor, validates
from sqlalchemy.schema import UniqueConstraint

from stalker import defaults
from stalker.db.declarative import Base
from stalker.models.mixins import ACLMixin
from stalker.models.entity import Entity
from stalker.log import logging_level

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging_level)


def group_finder(login, request):
    """Returns the group of the given login. The login name will be in
    'User:{login}' format.

    :param login: The login of the user, both '{login}' and
      'User:{login}' format is accepted.

    :param request: The Request object

    :return: Will return the groups of the user in ['Group:{group_name}']
      format.
    """

    if ':' in login:
        login = login.split(':')[1]

    # return the group of the given User object
    user_obj = User.query.filter_by(login=login).first()

    if user_obj:
        # just return the groups names if there is any group
        groups = user_obj.groups
        if len(groups):
            return map(lambda x: 'Group:' + x.name, groups)

    return []


class RootFactory(object):
    """The main purpose of having a root factory is to generate the objects
    used as the context by the request. But in our case it just used to
    determine the default ACLs.
    """

    @property
    def __acl__(self):
        # create the default acl and give admins all the permissions
        all_permissions = map(
            lambda x: x.action + '_' + x.class_name,
            Permission.query.all()
        )

        # start with default ACLs

        ACLs = [
            ('Allow', 'Group:' + defaults.admin_department_name,
             all_permissions),
            ('Allow', 'User:' + defaults.admin_name, all_permissions)
        ]

        # get all users and their ACLs
        all_users = User.query.all()
        for user in all_users:
            ACLs.extend(user.__acl__)

        # get all groups and their ACLs
        all_groups = Group.query.all()
        for group in all_groups:
            ACLs.extend(group.__acl__)

        return ACLs

    def __init__(self, request):
        pass


class Permission(Base):
    """A class to hold permissions.

    Permissions in Stalker defines what one can do or do not. A Permission
    instance is composed by three attributes; access, action and class_name.

    Permissions for all the classes in SOM are generally created by Stalker
    when initializing the database.

    If you created any custom classes to extend SOM you are also responsible to
    create the Permissions for it by calling :meth:`stalker.db.register` and
    passing your class to it. See the :mod:`stalker.db` documentation for
    details.

    :param str access: An Enum value which can have the one of the values of
      ``Allow`` or ``Deny``.

    :param str action: An Enum value from the list ['Create', 'Read', 'Update',
      'Delete', 'List']. Can not be None. The list can be changed from
      stalker.config.Config.default_actions.

    :param str class_name: The name of the class that this action is applied
      to. Can not be None or an empty string.

    Example: Let say that you want to create a Permission specifying a Group of
    Users are allowed to create Projects::

      from stalker import db
      from stalker.db.session import DBSession
      from stalker.models.auth import User, Group, Permission

      # first setup the db with the default database
      #
      # stalker.db.__init_db__ will create all the Actions possible with the
      # SOM classes automatically
      # 
      # What is left to you is to create the permissions
      setup.db()

      user1 = User(login='test_user1', password='1234')
      user2 = User(login='test_user2', password='1234')

      group1 = Group(name='users')
      group1.users = [user1, user2]

      # get the permissions for the Project class
      project_permissions = Permission.query\
          .filter(Actions.access='Allow')\
          .filter(Actions.action='Create')\
          .filter(Actions.class_name='Project')\
          .first()

      # now we have the permission specifying the allowance of creating a
      # Project

      # to make group1 users able to create a Project we simply add this
      # Permission to the groups permission attribute
      group1.permissions.append(permission)

      # and persist this information in the database
      DBSession.add(group)
      DBSession.commit()
    """
    __tablename__ = 'Permissions'
    __table_args__ = (
        UniqueConstraint('access', 'action', 'class_name'),
        {"extend_existing": True}
    )

    id = Column(Integer, primary_key=True)
    _access = Column('access', Enum('Allow', 'Deny', name='AccessNames'))
    _action = Column('action',
                     Enum(*defaults.actions, name='ActionNames'))
    _class_name = Column('class_name', String)

    def __init__(self, access, action, class_name):
        self._access = self._validate_access(access)
        self._action = self._validate_action(action)
        self._class_name = self._validate_class_name(class_name)

    def _validate_access(self, access):
        """validates the given access value
        """
        if not isinstance(access, (str, unicode)):
            raise TypeError('Permission.access should be an instance of str '
                            'or unicode not %s' % access.__class__.__name__)

        if access not in ['Allow', 'Deny']:
            raise ValueError('Permission.access should be "Allow" or "Deny"'
                             'not %s' % access)

        return access

    def _access_getter(self):
        """returns the _access value
        """
        return self._access

    access = synonym('_access', descriptor=property(_access_getter))

    def _validate_class_name(self, class_name):
        """validates the given class_name value
        """
        if not isinstance(class_name, (str, unicode)):
            raise TypeError('Permission.class_name should be an instance of '
                            'str or unicode not %s' %
                            class_name.__class__.__name__)

        return class_name

    def _class_name_getter(self):
        """returns the _class_name attribute value
        """
        return self._class_name

    class_name = synonym('_class_name',
                         descriptor=property(_class_name_getter))

    def _validate_action(self, action):
        """validates the given action value
        """

        if not isinstance(action, (str, unicode)):
            raise TypeError('Permission.action should be an instance of str '
                            'or unicode not %s' % action.__class__.__name__)

        if action not in defaults.actions:
            raise ValueError('Permission.action should be one of the values '
                             'of %s not %s' %
                             (defaults.actions, action))

        return action

    def _action_getter(self):
        """returns the _action value
        """
        return self._action

    action = synonym('_action', descriptor=property(_action_getter))

    def __eq__(self, other):
        """the equality of two Permissions
        """
        return isinstance(other, Permission) \
                   and other.access == self.access \
                   and other.action == self.action \
            and other.class_name == self.class_name

    def __ne__(self, other):
        """the inequality of two Permissions
        """
        return not self.__eq__(other)


class Group(Entity, ACLMixin):
    """Creates groups for users to be used in authorization system.

    A Group instance is nothing more than a list of :class:`.User`\ s created
    to be able to assign permissions in a group level.

    The Group class, as with the :class:`.User` class, is mixed with the
    :class:`~stalker.models.mixins.ACLMixin` which adds ability to hold
    :class:`~stalker.models.mixin.Permission` instances and serve ACLs to
    Pyramid.
    """

    # TODO: Update Group class documentation
    __auto_name__ = False
    __tablename__ = 'Groups'
    __mapper_args__ = {'polymorphic_identity': 'Group'}

    gid = Column("id", Integer, ForeignKey("Entities.id"),
                 primary_key=True)

    users = relationship(
        "User",
        secondary="User_Groups",
        back_populates="groups",
        doc="""Users in this group.

        Accepts:class:`.User` instance.
        """
    )

    def __init__(self, name='', users=None, **kwargs):
        if users is None:
            users = []
        kwargs.update({'name': name})
        super(Group, self).__init__(**kwargs)

        self.users = users

    @validates('users')
    def _validate_users(self, key, user):
        """validates the given user instance
        """
        if not isinstance(user, User):
            raise TypeError(
                'Group.users attribute must all be stalker.models.auth.User '
                'instances not %s' % user.__class__.__name__
            )

        return user


class User(Entity, ACLMixin):
    """The user class is designed to hold data about a User in the system.

    There are a couple of points to take your attention to:

    * The :attr:`.code` attribute is derived from the :attr:`.nice_name` as it
      is in a :class:`~stalker.models.entity.SimpleEntity`, but the
      :attr:`.nice_name` is derived from the :attr:`.login` instead of the
      :attr:`.name` attribute, so the :attr:`.code` of a :class:`.User` and a
      :class:`~stalker.models.entity.SimpleEntity` will be different then each
      other. The formatting of the :attr:`.code` attribute is as follows:

      * no underscore character is allowed, so while in the
        :class:`~stalker.models.entity.SimpleEntity` class the code could have
        underscores, in :class:`.User` class it is not
        allowed.
      * all the letters in the code will be converted to lower case.

      Other than these two new rules all the previous formatting rules from the
      :class:`~stalker.models.entity.SimpleEntity` are valid.

    .. note::
       .. versionadded 0.2.0: Task Watchers

       New to version 0.2.0 users can be assigned to a
       :class:`~stalker.models.task.Task` as a **Watcher**. Which can be used
       to inform the users in watchers list about the updates of certain Tasks.

    .. note::
       .. versionadded 0.2.0: Vacations

       It is now possible to define Vacations per user.

    .. note::
        .. deprecated 0.2.0: User.sequences_lead

        User.sequences_lead attribute has been deprecated. Use
        :attr:`.responsible_of` to query all the
        :class:`~stalker.models.task.Task`\ s and derivatives.

    :param email: holds the e-mail of the user, should be in [part1]@[part2]
      format

    :type email: str, unicode

    :param login: This is the login name of the user, it should be all lower
      case. Giving a string or unicode that has uppercase letters, it will be
      converted to lower case. It can not be an empty string or None and it can
      not contain any white space inside.

    :type login: str, unicode

    :param departments: It is the departments that the user is a part of. It
      should be a list of Department objects. One user can be listed in
      multiple departments.

    :type departments: list of :class:`~stalker.models.department.Department`\ s

    :param password: it is the password of the user, can contain any character.
      Stalker doesn't store the raw passwords of the users. To check a stored
      password with a raw password use :meth:`.check_password` and to set the
      password you can use the :attr:`.password` property directly.

    :type password: str, unicode

    :param groups: It is a list of :class:`.Group` instances that this user
      belongs to.

    :type groups: list of :class:`.Group`

    :param tasks: it is a list of Task objects which holds the tasks that this
      user has been assigned to

    :type tasks: list of :class:`~stalker.models.task.Task`\ s

    :param projects_lead: it is a list of Project objects that this user
      is the leader of, it is for back referencing purposes.

    :type projects_lead: list of :class:`~stalker.models.project.Project`\ s

    :param last_login: it is a datetime.datetime object holds the last login
      date of the user (not implemented yet)

    :type last_login: datetime.datetime
    """
    __auto_name__ = False
    __tablename__ = "Users"
    __mapper_args__ = {"polymorphic_identity": "User"}

    user_id = Column("id", Integer, ForeignKey("Entities.id"),
                     primary_key=True)

    departments = relationship(
        "Department",
        secondary='User_Departments',
        back_populates="members",
        doc="""A list of :class:`~stalker.models.department.Department`\ s that
        this user is a part of""",
    )

    email = Column(
        String(256),
        unique=True,
        nullable=False,
        doc="""email of the user, accepts strings or unicode"""
    )

    password = Column(
        String(256),
        nullable=False,
        doc="""The password of the user.

        It is scrambled before it is stored.
        """
    )

    last_login = Column(
        DateTime,
        doc="""The last login time of this user.

        It is an instance of datetime.datetime class."""
    )

    login = Column(
        String(256),
        nullable=False,
        doc="""The login name of the user.

        Can not be empty.
        """
    )

    groups = relationship(
        'Group',
        secondary='User_Groups',
        back_populates='users',
        doc="""Permission groups that this users is a member of.

        Accepts :class:`.Group` object.
        """
    )

    projects = relationship(
        'Project',
        secondary='Project_Users',
        back_populates='users'
    )

    projects_lead = relationship(
        "Project",
        primaryjoin="Projects.c.lead_id==Users.c.id",
        #uselist=True,
        back_populates="lead",
        doc=""":class:`~stalker.models.project.Project`\ s lead by this user.

        It is a list of :class:`~stalker.models.project.Project` instances.
        """
    )

    tasks = relationship(
        "Task",
        secondary="Task_Resources",
        back_populates="resources",
        doc=""":class:`~stalker.models.task.Task`\ s assigned to this user.

        It is a list of :class:`~stalker.models.task.Task` instances.
        """
    )

    watching = relationship(
        'Task',
        secondary='Task_Watchers',
        back_populates='watchers',
        doc=''':class:`~stalker.models.task.Tasks`\ s that this user is
        assigned as a watcher.

        It is a list of :class:`~stalker.models.task.Task` instances.
        '''
    )

    responsible_of = relationship(
        'Task',
        primaryjoin='Users.c.id==Tasks.c.responsible_id',
        back_populates='_responsible',
        uselist=True,
        doc='''A list of :class:`~stalker.models.task.Task` instances that this user is responsible of.
        '''
    )

    time_logs = relationship(
        "TimeLog",
        primaryjoin="TimeLogs.c.resource_id==Users.c.id",
        back_populates="resource",
        doc="""A list of :class:`~stalker.models.task.TimeLog` instances which
        holds the time logs created for this :class:`.User`.
        """
    )

    vacations = relationship(
        'Vacation',
        primaryjoin='Vacations.c.user_id==Users.c.id',
        back_populates='user',
        doc="""A list of :class:`~stalker.models.studio.Vacation` instances
        which holds the vacations created for this :class:`.User`
        """
    )

    def __init__(
            self,
            name=None,
            login=None,
            email=None,
            password=None,
            departments=None,
            groups=None,
            projects_lead=None,
            tasks=None,
            watching=None,
            # TODO: Watchlist should be a list of lists `WatchLists` so one can create task lists (playlists) for different tasks
            last_login=None,
            **kwargs
    ):
        kwargs['name'] = name

        super(User, self).__init__(**kwargs)

        self.login = login

        if departments is None:
            departments = []
        self.departments = departments

        self.email = email

        # to be able to mangle the password do it like this
        self.password = password

        if groups is None:
            groups = []
        self.groups = groups

        if projects_lead is None:
            projects_lead = []
        self.projects_lead = projects_lead

        if tasks is None:
            tasks = []

        self.tasks = tasks

        if watching is None:
            watching = []

        self.last_login = last_login

    # @reconstructor
    # def __init_on_load__(self):
    #     """initialized the instance variables when the instance created with
    #     SQLAlchemy
    #     """
    #     # call the Entity __init_on_load__
    #     super(User, self).__init_on_load__()

    def __repr__(self):
        """return the representation of the current User
        """
        return "<User (%s ('%s'))>" % (self.name, self.login)

    def __eq__(self, other):
        """the equality operator
        """
        return super(User, self).__eq__(other) and \
               isinstance(other, User) and \
               self.email == other.email and \
               self.login == other.login and \
               self.name == other.name

    def __ne__(self, other):
        """the inequality operator
        """
        return not self.__eq__(other)

    @validates("login")
    def _validate_login(self, key, login):
        """validates the given login value
        """
        if login is None:
            raise TypeError('%s.login can not be None' %
                            self.__class__.__name__)

        login = self._format_login(login)

        # raise a ValueError if the login is an empty string after formatting
        if login == '':
            raise ValueError('%s.login can not be an empty string' %
                             self.__class__.__name__)

        logger.debug("name out: %s" % login)

        return login

    @validates("departments")
    def _validate_department(self, key, department):
        """validates the given department value
        """
        from stalker.models.department import Department

        # check if it is instance of Department object
        if not isinstance(department, Department):
            raise TypeError("%s.department should be instance of "
                            "stalker.models.department.Department not %s" %
                            (self.__class__.__name__,
                             department.__class__.__name__))
        return department

    @validates("email")
    def _validate_email(self, key, email_in):
        """validates the given email value
        """
        # check if email_in is an instance of string or unicode
        if not isinstance(email_in, (str, unicode)):
            raise TypeError("%s.email should be an instance of string or "
                            "unicode not %s" %
                            (self.__class__.__name__,
                             email_in.__class__.__name__))
        return self._validate_email_format(email_in)

    def _validate_email_format(self, email_in):
        """formats the email
        """

        # split the mail from @ sign
        splits = email_in.split("@")
        len_splits = len(splits)

        # there should be one and only one @ sign
        if len_splits > 2:
            raise ValueError("check the %s.email formatting, there are more "
                             "than one @ sign" % self.__class__.__name__)

        if len_splits < 2:
            raise ValueError("check the %s.email formatting, there are no @ "
                             "sign" % self.__class__.__name__)

        if splits[0] == "":
            raise ValueError("check the %s.email formatting, the name part "
                             "is missing" % self.__class__.__name__)

        if splits[1] == "":
            raise ValueError("check the %s.email formatting, the domain part "
                             "is missing" % self.__class__.__name__)

        return email_in

    @validates("last_login")
    def _validate_last_login(self, key, last_login_in):
        """validates the given last_login argument
        """
        if not isinstance(last_login_in, datetime.datetime) and \
                        last_login_in is not None:
            raise TypeError("%s.last_login should be an instance of "
                            "datetime.datetime or None not %s" %
                            (self.__class__.__name__,
                             last_login_in.__class__.__name__))
        return last_login_in

    def _format_login(self, login):
        """formats the given login value
        """
        # be sure it is a string
        login = str(login)

        # strip white spaces from start and end
        login = login.strip()

        # remove all the spaces
        login = login.replace(" ", "")

        # make it lowercase
        login = login.lower()

        # remove any illegal characters
        login = re.sub("[^\\(a-zA-Z0-9)]+", "", login)

        # remove any number at the beginning
        login = re.sub("^[0-9]+", "", login)

        return login

    @validates("password")
    def _validate_password(self, key, password_in):
        """validates the given password
        """
        if password_in is None:
            raise TypeError("%s.password cannot be None" %
                            self.__class__.__name__)

        if password_in == "":
            raise ValueError("raw_password can not be empty string")

        # mangle the password
        return base64.encodestring(password_in)

    def check_password(self, raw_password):
        """Checks the given raw_password.

        Checks the given raw_password with the current Users objects encrypted
        password.

        Checks the given raw password with the given encrypted password.
        Handles the encryption process behind the scene.
        """
        return self.password == base64.encodestring(str(raw_password))

    @validates("groups")
    def _validate_groups(self, key, group):
        """check the given group
        """
        if not isinstance(group, Group):
            raise TypeError(
                "any group in %s.groups should be an instance of"
                "stalker.models.auth.Group not %s" %
                (self.__class__.__name__, group.__class__.__name__)
            )

        return group

    @validates("projects_lead")
    def _validate_projects_lead(self, key, project):
        """validates the given projects_lead attribute
        """
        from stalker.models.project import Project

        if not isinstance(project, Project):
            raise TypeError(
                "any element in %s.projects_lead should be a"
                "stalker.models.project.Project instance not %s" %
                (self.__class__.__name__, project.__class__.__name__)
            )
        return project

    @validates("tasks")
    def _validate_tasks(self, key, task):
        """validates the given tasks attribute
        """
        from stalker.models.task import Task

        if not isinstance(task, Task):
            raise TypeError(
                "any element in %s.tasks should be an instance of "
                "stalker.models.task.Task not %s" %
                (self.__class__.__name__, task.__class__.__name__)
            )
        return task

    @validates("watching")
    def _validate_watching(self, key, task):
        """validates the given watching attribute
        """
        from stalker.models.task import Task

        if not isinstance(task, Task):
            raise TypeError(
                "any element in %s.watching should be an instance of "
                "stalker.models.task.Task not %s" %
                (self.__class__.__name__, task.__class__.__name__)
            )
        return task

    @validates('projects')
    def _validate_projects(self, key, project):
        """validates the given project instance
        """
        from stalker import Project

        if not isinstance(project, Project):
            raise TypeError('%s.projects should a list of '
                            'stalker.models.project.Project instances, not '
                            '%s' % (self.__class__.__name__,
                                    project.__class__.__name__))
        return project

    @validates('vacations')
    def _validate_vacations(self, key, vacation):
        """validates the given vacation value
        """
        from stalker.models.studio import Vacation

        if not isinstance(vacation, Vacation):
            raise TypeError("Every member of %s.vacations should be an "
                            "instance of stalker.models.studio.Vacation, "
                            "not %s" %
                            (self.__class__.__name__,
                             vacation.__class__.__name__))

        return vacation

    @property
    def tickets(self):
        """The list of :class:`~stalker.models.ticket.Ticket`\ s that this user has.

        returns a list of :class:`~stalker.models.ticket.Ticket` instances
        which are derived from the :class:`~stalker.models.task.Task`\ s that
        this user is assigned to
        (Stalker checks the related :class:`~stalker.models.version.Version`
        instances and then the `~stalker.models.ticket.Ticket` instances
        assigned to the Versions.).
        """
        # do it with sqlalchemy
        from stalker import Ticket

        return Ticket.query \
            .filter(Ticket.owner == self) \
            .all()

    @property
    def open_tickets(self):
        """The list of open :class:`~stalker.models.ticket.Ticket`\ s that this user has.

        returns a list of :class:`~stalker.models.ticket.Ticket` instances
        which has a status of `Open` and are derived from the
        :class:`~stalker.models.task.Task`\ s that this user is assigned to
        (Stalker checks the related :class:`~stalker.models.version.Version`
        instances and then the `~stalker.models.ticket.Ticket` instances
        assigned to the Version and has a status of `Open`.).
        """
        # do it with sqlalchemy
        from stalker import Ticket, Status

        return Ticket.query \
            .filter(Ticket.owner == self) \
            .join(Ticket.status) \
            .filter(Status.code != 'CLOSED') \
            .all()

    @property
    def to_tjp(self):
        """outputs a TaskJuggler formatted string
        """
        from jinja2 import Template

        temp = Template(defaults.tjp_user_template)
        return temp.render({'user': self})


class LocalSession(object):
    """A simple temporary session object which simple stores session data.

    This class will later be removed, it is here because we need a login window
    for the Qt user interfaces.

    On initialize it will load the SessionData from the users .strc folder
    """

    def __init__(self):
        self.logged_in_user_id = None
        self.valid_to = None
        self.session_data = None
        self.load()

    def load(self):
        """loads the data from the saved local session
        """
        try:
            with open(LocalSession.session_file_full_path(), 'r') as s:
                # try:
                unpickled_object = pickle.load(s)
                if unpickled_object.valid_to > datetime.datetime.now():
                    # fill __dict__ with the loaded one
                    self.__dict__ = unpickled_object.__dict__
        except IOError:
            pass

    @property
    def logged_in_user(self):
        """returns the logged in user
        """
        return User.query.filter_by(id=self.logged_in_user_id).first()

    def store_user(self, user):
        """stores the given user instance

        :param user: The user instance. 
        """
        if user:
            self.logged_in_user_id = user.id

    def save(self):
        """remembers the data in user local file system
        """
        self.valid_to = datetime.datetime.now() + datetime.timedelta(10)
        # serialize self
        dumped_data = pickle.dumps(self)
        logger.debug('dumped session data : %s' % dumped_data)
        self._write_data(dumped_data)

    def delete(self):
        """removes the cache file
        """
        try:
            os.remove(self.session_file_full_path())
        except OSError:
            pass

    @classmethod
    def session_file_full_path(cls):
        """
        :return str: the session file full path
        """
        return os.path.normpath(os.path.join(
            defaults.local_storage_path,
            defaults.local_session_data_file_name
        ))

    def _write_data(self, data):
        """Writes the given data to the local session file

        :param data: the data to be written (generally serialized LocalSession
          class itself)
        """
        file_full_path = self.session_file_full_path()

        # create the path first
        file_path = os.path.dirname(file_full_path)
        try:
            os.makedirs(file_path)
        except OSError:
            # dir exists
            pass
        finally:
            with open(file_full_path, 'w') as data_file:
                data_file.writelines(data)


# USER_PERMISSIONGROUPS
User_Groups = Table(
    "User_Groups", Base.metadata,
    Column("uid", Integer, ForeignKey("Users.id"), primary_key=True),
    Column("gid", Integer, ForeignKey("Groups.id"), primary_key=True)
)

# USER_DEPARTMENTS
User_Departments = Table(
    'User_Departments', Base.metadata,
    Column('uid', Integer, ForeignKey('Users.id'), primary_key=True),
    Column('did', Integer, ForeignKey('Departments.id'), primary_key=True)
)



# -*- coding: utf-8 -*-
# Stalker a Production Asset Management System
# Copyright (C) 2009-2013 Erkan Ozgur Yilmaz
#
# This file is part of Stalker.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import unittest2
import datetime
from stalker import (Project, Repository, Status, StatusList, Task, TimeLog,
                     User)
from stalker.exceptions import OverBookedError


class TimeLogTester(unittest2.TestCase):
    """tests the TimeLog class
    """

    def setUp(self):
        """setup the test
        """
        # create a resource
        self.test_resource1 = User(
            name="User1",
            login="user1",
            email="user1@users.com",
            password="1234",
        )

        self.test_resource2 = User(
            name="User2",
            login="user2",
            email="user2@users.com",
            password="1234"
        )

        self.test_repo = Repository(name="test repository")

        # create a Project
        self.test_status1 = Status(name="Status1", code="STS1")
        self.test_status2 = Status(name="Status2", code="STS2")
        self.test_status3 = Status(name="Status3", code="STS3")

        self.test_project_status_list = StatusList(
            name="Project Statuses",
            statuses=[self.test_status1],
            target_entity_type=Project
        )

        self.test_task_status_list = StatusList(
            name="Task Statuses",
            statuses=[self.test_status1, self.test_status2, self.test_status3],
            target_entity_type=Task
        )

        self.test_project = Project(
            name="test project",
            code='tp',
            repository=self.test_repo,
            status_list=self.test_project_status_list
        )

        # create a Task
        self.test_task = Task(
            name="test task",
            project=self.test_project,
            status_list=self.test_task_status_list
        )

        self.kwargs = {
            "name": "test time_log",
            "task": self.test_task,
            "resource": self.test_resource1,
            "start": datetime.datetime(2013, 3, 22, 1, 0),
            "duration": datetime.timedelta(10)
        }

        # create a TimeLog
        # and test it
        self.test_time_log = TimeLog(**self.kwargs)

    def test___auto_name__class_attribute_is_set_to_True(self):
        """testing if the __auto_name__ class attribute is set to True for
        TimeLog class
        """
        self.assertTrue(TimeLog.__auto_name__)

    def test_task_argument_is_Skipped(self):
        """testing if a TypeError will be raised when the task argument is
        skipped
        """
        self.kwargs.pop("task")
        self.assertRaises(TypeError, TimeLog, **self.kwargs)

    def test_task_argument_is_None(self):
        """testing if a TypeError will be raised when the task argument is None
        """
        self.kwargs["task"] = None
        self.assertRaises(TypeError, TimeLog, **self.kwargs)

    def test_task_attribute_is_None(self):
        """testing if a TypeError will be raised when the task attribute i
        None
        """
        self.assertRaises(TypeError, setattr, self.test_time_log, "task", None)

    def test_task_argument_is_not_a_Task_instance(self):
        """testing if a TypeError will be raised when the task argument is not
        a stalker.models.task.Task instance
        """
        self.kwargs["task"] = "this is a task"
        self.assertRaises(TypeError, TimeLog, **self.kwargs)

    def test_task_attribute_is_not_a_Task_instance(self):
        """testing if a TypeError will be raised when the task attribute is not
        a stalker.models.task.Task instance
        """
        self.assertRaises(TypeError, setattr, self.test_time_log, "task",
                          "this is a task")

    def test_task_attribute_is_working_properly(self):
        """testing if the task attribute is working properly
        """
        new_task = Task(
            name="Test task 2",
            project=self.test_project,
            status_list=self.test_task_status_list,
            resources=[self.test_resource1],
        )
        self.assertNotEqual(self.test_time_log.task, new_task)
        self.test_time_log.task = new_task
        self.assertEqual(self.test_time_log.task, new_task)

    def test_task_argument_updates_backref(self):
        """testing if the Task given with the task argument is updated correctly
        with the current TimeLog instance is listed in the time_logs attribute of
        the Task
        """
        new_task = Task(
            name="Test Task 3",
            project=self.test_project,
            status_list=self.test_task_status_list,
            resources=[self.test_resource1],
        )

        # now create a new time_log for the new task
        self.kwargs["task"] = new_task
        self.kwargs["start"] = self.kwargs["start"] + \
                               self.kwargs["duration"] + \
                               datetime.timedelta(120)
        new_time_log = TimeLog(**self.kwargs)

        # now check if the new_time_log is in task.time_logs
        self.assertIn(new_time_log, new_task.time_logs)

    def test_task_attribute_updates_backref(self):
        """testing if the Task given with the task attribute is updated
        correctly with the current TimeLog instance is listed in the time_logs
        attribute of the Task
        """
        new_task = Task(
            name="Test Task 3",
            project=self.test_project,
            status_list=self.test_task_status_list,
            resources=[self.test_resource1],
        )

        self.test_time_log.task = new_task
        self.assertIn(self.test_time_log, new_task.time_logs)

    def test_resource_argument_is_skipped(self):
        """testing if a TypeError will be raised when the resource argument is
        skipped
        """
        self.kwargs.pop("resource")
        self.assertRaises(TypeError, TimeLog, **self.kwargs)

    def test_resource_argument_is_None(self):
        """testing if a TypeError will be raised when the resource argument is
        None
        """
        self.kwargs["resource"] = None
        self.assertRaises(TypeError, TimeLog, **self.kwargs)

    def test_resource_attribute_is_None(self):
        """testing if a TypeError will be raised when the resource attribute is
        set to None
        """
        self.assertRaises(TypeError, setattr, self.test_time_log, "resource",
                          None)

    def test_resource_argument_is_not_a_User_instance(self):
        """testing if a TypeError will be raised when the resource argument is
        not a stalker.models.user.User instance
        """
        self.kwargs["resource"] = "This is a resource"
        self.assertRaises(TypeError, TimeLog, **self.kwargs)

    def test_resource_attribute_is_not_a_User_instance(self):
        """testing if a TypeError will be raised when the resource attribute is
        set to a value other than a stalker.models.user.User instance
        """
        self.assertRaises(TypeError, setattr, self.test_time_log, "resource",
                          "this is a resource")

    def test_resource_attribute_is_working_properly(self):
        """testing if the resource attribute is working properly
        """
        new_resource = User(
            name="Test Resource",
            login="test resource 2",
            email="test@resource2.com",
            password="1234",
        )

        self.assertNotEqual(self.test_time_log.resource, new_resource)
        self.test_time_log.resource = new_resource
        self.assertEqual(self.test_time_log.resource, new_resource)

    def test_resource_argument_updates_backref(self):
        """testing if the User instance given with the resource argument is
        updated with the current TimeLog is listed in the time_logs attribute of
        the User instance
        """
        new_resource = User(
            name="Test Resource",
            login="test resource 2",
            email="test@resource2.com",
            password="1234",
        )

        self.kwargs["resource"] = new_resource
        new_time_log = TimeLog(**self.kwargs)

        self.assertEqual(new_time_log.resource, new_resource)

    def test_resource_attribute_updates_backref(self):
        """testing if the User instance given with the resource attribute is
        updated with the current TimeLog is listed in the time_logs attribute of
        the User instance
        """
        new_resource = User(
            name="Test Resource",
            login="test resource 2",
            email="test@resource2.com",
            password="1234",
        )

        self.assertNotEqual(self.test_time_log.resource, new_resource)
        self.test_time_log.resource = new_resource
        self.assertEqual(self.test_time_log.resource, new_resource)

    def test_ScheduleMixin_initialization(self):
        """testing if the ScheduleMixin part is initialized correctly
        """
        # it should have schedule attributes
        self.assertEqual(self.test_time_log.start,
                         self.kwargs["start"])
        self.assertEqual(self.test_time_log.duration, self.kwargs["duration"])

        self.test_time_log.start = datetime.datetime(2013, 3, 22, 4, 0)
        self.test_time_log.end = self.test_time_log.start + \
                                 datetime.timedelta(10)
        self.assertEqual(self.test_time_log.duration, datetime.timedelta(10))

    def test_OverbookedError_1(self):
        """testing if a OverBookedError will be raised when the resource 
        is already booked for the given time period.
        
        Simple case diagram:
        #####
        #####
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0),
        self.kwargs["duration"] = datetime.timedelta(10)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2
        self.kwargs["name"] = "time_log2"
        
        self.assertRaises(OverBookedError, TimeLog, **self.kwargs)

    def test_OverbookedError_2(self):
        """testing if a OverBookedError will be raised when the resource 
        is already booked for the given time period.
        
        Simple case diagram:
        #######
        #####
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0),
        self.kwargs["duration"] = datetime.timedelta(10)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2
        self.kwargs["name"] = "time_log2"
        self.kwargs["duration"] = datetime.timedelta(8)

        self.assertRaises(OverBookedError, TimeLog, **self.kwargs)

    def test_OverbookedError_3(self):
        """testing if a OverBookedError will be raised when the resource 
        is already booked for the given time period.
        
        Simple case diagram:
        #####
        #######
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0),
        self.kwargs["duration"] = datetime.timedelta(8)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2        
        self.kwargs["name"] = "time_log2"
        self.kwargs["duration"] = datetime.timedelta(10)

        self.assertRaises(OverBookedError, TimeLog, **self.kwargs)
    
    def test_OverbookedError_4(self):
        """testing if a OverBookedError will be raised when the resource 
        is already booked for the given time period.
        
        Simple case diagram:        
        #######
          #####
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0) - \
                               datetime.timedelta(2)
        self.kwargs["duration"] = datetime.timedelta(12)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2
        self.kwargs["name"] = "time_log2"
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0)
        self.kwargs["duration"] = datetime.timedelta(10)

        self.assertRaises(OverBookedError, TimeLog, **self.kwargs)

    def test_OverbookedError_5(self):
        """testing if a OverBookedError will be raised when the resource 
        is already booked for the given time period.
        
        Simple case diagram:
          #####
        #######
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0)
        self.kwargs["duration"] = datetime.timedelta(10)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2
        self.kwargs["name"] = "time_log2"
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0) - \
                               datetime.timedelta(2)
        self.kwargs["duration"] = datetime.timedelta(12)

        self.assertRaises(OverBookedError, TimeLog, **self.kwargs)

    def test_OverbookedError_6(self):
        """testing if a OverBookedError will be raised when the resource 
        is already booked for the given time period.
        
        Simple case diagram:
          #######
        #######
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0)
        self.kwargs["duration"] = datetime.timedelta(15)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2
        self.kwargs["name"] = "time_log2"
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0) - \
                               datetime.timedelta(5)
        self.kwargs["duration"] = datetime.timedelta(15)

        self.assertRaises(OverBookedError, TimeLog, **self.kwargs)

    def test_OverbookedError_7(self):
        """testing if a OverBookedError will be raised when the resource 
        is already booked for the given time period.
        
        Simple case diagram:
        #######
          #######
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0) - \
                               datetime.timedelta(5)
        self.kwargs["duration"] = datetime.timedelta(15)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2        
        self.kwargs["name"] = "time_log2"
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0)
        self.kwargs["duration"] = datetime.timedelta(15)

        import warnings

        self.assertRaises(OverBookedError, TimeLog, **self.kwargs)

    def test_OverbookedError_8(self):
        """testing if no OverBookedError will be raised when the resource 
        is not already booked for the given time period.
        
        Simple case diagram:
        #######
                 #######
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0)
        self.kwargs["duration"] = datetime.timedelta(5)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2
        self.kwargs["name"] = "time_log2"
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0) + \
                               datetime.timedelta(20)
        # no warning
        time_log2 = TimeLog(**self.kwargs)

    def test_OverbookedError_9(self):
        """testing if no OverBookedError will be raised when the resource 
        is not already booked for the given time period.
        
        Simple case diagram:
                 #######
        #######
        """
        # time_log1
        self.kwargs["name"] = "time_log1"
        self.kwargs["resource"] = self.test_resource2
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0) + \
                               datetime.timedelta(20)
        self.kwargs["duration"] = datetime.timedelta(5)

        time_log1 = TimeLog(**self.kwargs)

        # time_log2
        self.kwargs["name"] = "time_log2"
        self.kwargs["start"] = datetime.datetime(2013, 3, 22, 4, 0)

        # no warning
        time_log2 = TimeLog(**self.kwargs)
    
    def test_OverbookedError_10(self):
        """testing i no OverBookedError will be raised for the same TimeLog
        instance
        """
        # time_log1
        self.kwargs['name'] = 'time_log1'
        self.kwargs['resource'] = self.test_resource2
        self.kwargs['start'] = datetime.datetime(2013, 5, 6, 14, 0)
        self.kwargs['duration'] = datetime.timedelta(20)
        
        time_log1 = TimeLog(**self.kwargs)
        
        # no warning
        self.test_resource2.time_logs.append(time_log1)

    def test_time_log_extends_effort_of_task(self):
        """testing if the TimeLog will expand the Task.schedule_timing if it is
        getting longer than Task.total_logged_seconds
        """
        task1 = Task(
            name='Test Task 1',
            schedule_timing=10,
            schedule_unit='h',
            schedule_model='effort',
            resource=self.test_resource1,
            status_list=self.test_task_status_list,
            project=self.test_project
        )
        
        # check everything is initialized as they should be
        self.assertEqual(task1.total_logged_seconds, 0)
        self.assertEqual(task1.remaining_seconds, 36000)
        
        dt = datetime.datetime
        td = datetime.timedelta
        
        # now create time log for the task
        timeLog1 = TimeLog(
            task=task1,
            resource=self.test_resource1,
            start=dt(2013, 5, 2, 15, 0),
            duration=td(hours=5)
        )
        
        # now check if the remaining seconds is correctly calculated
        self.assertEqual(
            task1.remaining_seconds, 18000
        )
        
        # and the schedule_timing is not expanded
        self.assertEqual(
            task1.schedule_timing,
            10
        )
        
        # now add a new timeLog which expands the Task
        timeLog2 = TimeLog(
            task=task1,
            resource=self.test_resource1,
            start=dt(2013, 5, 2, 20, 0),
            duration=td(hours=8)
        )
        
        # check if the task schedule_timing is correctly expanded to 13 hours
        self.assertEqual(
            task1.schedule_timing,
            13
        )
    
    def test_time_log_extends_effort_of_task_with_different_time_unit(self):
        """testing if the TimeLog will expand the Task.schedule_timing if it is
        getting longer than Task.total_logged_seconds and can work with
        different time units
        """
        from stalker import defaults
        defaults.daily_working_hours = 10

        task1 = Task(
            name='Test Task 1',
            schedule_timing=2,
            schedule_unit='d',
            schedule_model='effort',
            resource=self.test_resource1,
            status_list=self.test_task_status_list,
            project=self.test_project
        )
        
        # check everything is initialized as they should be
        self.assertEqual(task1.total_logged_seconds, 0)
        
        # there are no studio created so it should use the defaults
        self.assertEqual(task1.remaining_seconds,
                         2 * 10 * 60 * 60)
        
        dt = datetime.datetime
        td = datetime.timedelta
        
        # now create time log for the task
        timeLog1 = TimeLog(
            task=task1,
            resource=self.test_resource1,
            start=dt(2013, 5, 2, 10, 0),
            duration=td(hours=10)
        )
        
        # now check if the remaining seconds is correctly calculated
        self.assertEqual(
            task1.remaining_seconds, 10 * 60 * 60
        )
        
        # and the schedule_timing is not expanded
        self.assertEqual(
            task1.schedule_timing,
            2
        )
        
        # now add a new timeLog which expands the Task
        timeLog2 = TimeLog(
            task=task1,
            resource=self.test_resource1,
            start=dt(2013, 5, 2, 20, 0),
            duration=td(hours=15)
        )
        
        # check if the task schedule_timing is correctly expanded to 2.5 days
        self.assertEqual(
            task1.schedule_timing,
            2.5
        )
        
